from collections.abc import Generator

from draive import MultimodalContent, async_noop, auto_retry, cache, ctx, tool
from pytest import mark, raises


class FakeException(Exception):
    pass


@mark.asyncio
@ctx.wrap("test")
async def test_call_returns_result():
    executions: int = 0

    @tool
    async def compute(value: int) -> int:
        nonlocal executions
        executions += 1
        return value

    assert await compute(value=42) == 42
    assert executions == 1


@mark.asyncio
@ctx.wrap("test")
async def test_call_fails_on_error():
    executions: int = 0

    @tool
    async def compute(value: int) -> int:
        nonlocal executions
        executions += 1
        raise FakeException()

    with raises(FakeException):
        await compute(value=42)
    assert executions == 1


@mark.asyncio
@ctx.wrap("test")
async def test_toolbox_call_returns_multimodal_content():
    executions: int = 0

    @tool
    async def compute(value: int) -> int:
        nonlocal executions
        executions += 1
        return value

    assert await compute._toolbox_call(
        "call_id",
        arguments={
            "value": 42,
        },
        report_status=async_noop,
    ) == MultimodalContent.of("42")
    assert executions == 1


@mark.asyncio
@ctx.wrap("test")
async def test_toolbox_call_returns_custom_content():
    executions: int = 0

    def custom_format(value: int) -> str:
        return f"Value:{value}"

    @tool(format_result=custom_format)
    async def compute(value: int) -> int:
        nonlocal executions
        executions += 1
        return value

    assert await compute._toolbox_call(
        "call_id",
        arguments={
            "value": 42,
        },
        report_status=async_noop,
    ) == MultimodalContent.of("Value:42")
    assert executions == 1


@mark.asyncio
@ctx.wrap("test")
async def test_retries_with_auto_retry():
    executions: int = 0

    @tool
    @auto_retry
    async def compute(value: int) -> int:
        nonlocal executions
        executions += 1
        if executions == 1:
            raise FakeException()
        else:
            return value

    assert await compute(value=42) == 42
    assert executions == 2


@mark.asyncio
@ctx.wrap("test")
async def test_returns_cached_with_cache():
    def fake_random() -> Generator[int, None, None]:
        yield from range(0, 65536)

    @tool
    @cache
    async def compute(value: int) -> int:
        return next(fake_random())

    expected: int = await compute(value=42)
    assert await compute(value=42) == expected



from collections.abc import Iterable
from typing import Self, cast, final

from draive.parameters import ParametrizedData
from draive.scope.errors import MissingScopeState
from draive.utils import MISSING, Missing, not_missing

__all__ = [
    "ScopeState",
]


@final
class ScopeState:
    def __init__(
        self,
        *state: ParametrizedData,
    ) -> None:
        self._state: dict[type[ParametrizedData], ParametrizedData] = {
            type(element): element for element in state
        }

    def state[State_T: ParametrizedData](
        self,
        state: type[State_T],
        /,
        default: State_T | Missing = MISSING,
    ) -> State_T:
        if state in self._state:
            return cast(State_T, self._state[state])

        elif not_missing(default):
            return default

        else:
            try:
                initialized: State_T = state()
                self._state[state] = initialized
                return initialized

            except (TypeError, AttributeError) as exc:
                raise MissingScopeState(
                    f"{state.__qualname__} is not defined in the current scope"
                    " and failed to provide a default value"
                ) from exc

    def updated(
        self,
        state: Iterable[ParametrizedData] | None,
    ) -> Self:
        if state:
            return self.__class__(
                *[
                    *self._state.values(),
                    *state,
                ]
            )
        else:
            return self



from typing import Any, Literal, NotRequired, Required, TypedDict

from draive.parameters import DataModel

__all__ = [
    "UsageInfo",
    "EmbeddingObject",
    "EmbeddingResponse",
    "ChatFunctionCall",
    "ChatToolCallRequest",
    "ChatMessage",
    "ChatFunctionCallResponse",
    "ChatToolCallResponse",
    "ChatMessageResponse",
    "ChatCompletionResponseChoice",
    "ChatCompletionResponse",
    "ChatDeltaMessageResponse",
    "ChatCompletionResponseStreamChoice",
    "ChatCompletionStreamResponse",
]


class UsageInfo(DataModel):
    prompt_tokens: int
    total_tokens: int
    completion_tokens: int | None = None


class EmbeddingObject(DataModel):
    object: str
    embedding: list[float]
    index: int


class EmbeddingResponse(DataModel):
    id: str
    object: str
    data: list[EmbeddingObject]
    model: str
    usage: UsageInfo


class ChatFunctionCall(TypedDict, total=False):
    name: Required[str]
    arguments: Required[str]


class ChatToolCallRequest(TypedDict, total=False):
    id: Required[str]
    type: Required[Literal["function"]]
    function: Required[ChatFunctionCall]


class ChatMessage(TypedDict, total=False):
    role: Required[str]
    content: Required[str | list[str]]
    name: NotRequired[str]
    tool_calls: NotRequired[list[ChatToolCallRequest]]


class ChatFunctionCallResponse(DataModel):
    name: str
    arguments: dict[str, Any] | str


class ChatToolCallResponse(DataModel):
    id: str
    type: Literal["function"]
    function: ChatFunctionCallResponse


class ChatDeltaMessageResponse(DataModel):
    role: str | None = None
    content: str | None = None
    tool_calls: list[ChatToolCallResponse] | None = None


class ChatCompletionResponseStreamChoice(DataModel):
    index: int
    delta: ChatDeltaMessageResponse
    finish_reason: Literal["stop", "length", "error", "tool_calls"] | None = None


class ChatCompletionStreamResponse(DataModel):
    id: str
    model: str
    choices: list[ChatCompletionResponseStreamChoice]
    created: int | None = None
    usage: UsageInfo | None = None


class ChatMessageResponse(DataModel):
    role: str
    content: list[str] | str | None = None
    tool_calls: list[ChatToolCallResponse] | None = None


class ChatCompletionResponseChoice(DataModel):
    index: int
    message: ChatMessageResponse
    finish_reason: Literal["stop", "length", "error", "tool_calls"] | None = None


class ChatCompletionResponse(DataModel):
    id: str
    object: str
    created: int
    model: str
    choices: list[ChatCompletionResponseChoice]
    usage: UsageInfo



import types
import typing
from collections.abc import Callable, Iterator
from types import GenericAlias, NoneType, UnionType
from typing import (
    Any,
    ForwardRef,
    ParamSpec,
    TypeAliasType,
    TypeVar,
    TypeVarTuple,
    _UnionGenericAlias,  # pyright: ignore[reportAttributeAccessIssue, reportUnknownVariableType]
    _UnpackGenericAlias,  # pyright: ignore[reportAttributeAccessIssue, reportUnknownVariableType]
    get_args,
    get_origin,
    get_type_hints,
)

import draive.utils.missing as draive_missing

__all__ = [
    "object_annotations",
    "resolve_annotation",
    "allows_missing",
    "ParameterDefaultFactory",
]


type ParameterDefaultFactory[Value] = Callable[[], Value]


def object_annotations(
    annotation: Any,
    /,
    globalns: dict[str, Any] | None,
    localns: dict[str, Any] | None,
) -> dict[str, Any]:
    return get_type_hints(
        annotation,
        globalns=globalns,
        localns=localns,
        include_extras=False,
    )


def allows_missing(
    annotation: Any,
    /,
    type_arguments: dict[str, Any],
    globalns: dict[str, Any] | None,
    localns: dict[str, Any] | None,
) -> bool:
    resolved_type, resolved_args = resolve_annotation(
        annotation,
        type_arguments=type_arguments,
        globalns=globalns,
        localns=localns,
    )

    match resolved_type:
        case draive_missing.Missing:
            return True

        case types.UnionType:
            return draive_missing.Missing in resolved_args

        case _:
            return False


def resolve_annotation(  # noqa: PLR0911, C901, PLR0912
    annotation: Any,
    /,
    type_arguments: dict[str, Any],
    globalns: dict[str, Any] | None,
    localns: dict[str, Any] | None,
) -> tuple[Any, tuple[Any, ...]]:
    match annotation:
        case NoneType() | None:
            return (
                NoneType,
                (),
            )

        case type() as type_annotation:
            return (
                get_origin(type_annotation) or type_annotation,
                get_args(type_annotation),
            )

        case UnionType() as union_annotation:
            return (
                UnionType,  # pyright: ignore[reportReturnType]
                get_args(union_annotation),
            )

        case _UnionGenericAlias() as union_alias_annotation:  # pyright: ignore[reportUnknownVariableType]
            return (
                UnionType,  # pyright: ignore[reportReturnType]
                get_args(union_alias_annotation),
            )

        case TypeAliasType() as alias_annotation:
            return resolve_annotation(
                alias_annotation.__value__,
                type_arguments=type_arguments,
                globalns=globalns,
                localns=localns,
            )

        case GenericAlias() as generic_annotation:
            match generic_annotation.__origin__:
                case TypeAliasType() as alias:  # pyright: ignore[reportUnnecessaryComparison]
                    resolved_args: Iterator[Any] = generic_annotation.__args__.__iter__()
                    # TODO: only basic cases are covered currently
                    merged_arguments: list[Any] = []
                    alias_value: Any = alias.__value__
                    for arg in get_args(alias_value):
                        match arg:
                            case TypeVar():
                                merged_arguments.append(next(resolved_args))

                            case _UnpackGenericAlias():
                                merged_arguments.extend(resolved_args)

                            case other:
                                merged_arguments.append(other)

                    return (
                        get_origin(alias_value) or alias_value,
                        tuple(merged_arguments),
                    )

                case parametrized if hasattr(parametrized, "__PARAMETERS__"):
                    resolved_parameters: tuple[Any, ...] = tuple(
                        type_arguments.get(
                            arg.__name__,
                            arg.__bound__ or Any,
                        )
                        if isinstance(arg, TypeVar)
                        else arg
                        for arg in generic_annotation.__args__
                    )
                    # ParametrizedData subtypes always have __class_getitem__
                    resolved_type: type[Any] = parametrized[resolved_parameters]  # pyright: ignore[reportIndexIssue, reportUnknownVariableType]

                    return (
                        resolved_type,  # pyright: ignore[reportUnknownVariableType]
                        get_args(resolved_type),
                    )

                case origin:
                    return (
                        origin,
                        generic_annotation.__args__,
                    )

        case TypeVar() as variable_annotation:
            parameter_type: Any = type_arguments.get(
                variable_annotation.__name__,
                variable_annotation.__bound__ or Any,
            )

            match parameter_type:
                case typing.TypeVar():
                    return resolve_annotation(
                        variable_annotation.__bound__ or Any,
                        type_arguments=type_arguments,
                        globalns=globalns,
                        localns=localns,
                    )

                case other:
                    return resolve_annotation(
                        other,
                        type_arguments=type_arguments,
                        globalns=globalns,
                        localns=localns,
                    )

        case ParamSpec():
            raise NotImplementedError("ParamSpec type parameter is not supported yet")

        case TypeVarTuple():
            raise NotImplementedError("TypeVarTuple type parameter is not supported yet")

        case str() as str_annotation:
            return resolve_annotation(
                ForwardRef(str_annotation)._evaluate(  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]
                    globalns=globalns,
                    localns=localns,
                    recursive_guard=frozenset(),
                ),
                type_arguments=type_arguments,
                globalns=globalns,
                localns=localns,
            )

        case ForwardRef() as ref_annotation:
            return resolve_annotation(
                ref_annotation._evaluate(  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]
                    globalns=globalns,
                    localns=localns,
                    recursive_guard=frozenset(),
                ),
                type_arguments=type_arguments,
                globalns=globalns,
                localns=localns,
            )

        case other:
            match get_origin(other):
                case typing.Annotated | typing.Final | typing.Required | typing.NotRequired:
                    return resolve_annotation(
                        get_args(other)[0],
                        type_arguments=type_arguments,
                        globalns=globalns,
                        localns=localns,
                    )

                case None:
                    return (
                        other,
                        get_args(other),
                    )

                case other_origin:
                    return (
                        other_origin,
                        get_args(other),
                    )



from draive.helpers.retry import auto_retry
from draive.helpers.trace import traced
from draive.helpers.volatile_index import VolatileVectorIndex
from draive.helpers.volatile_memory import (
    ConstantMemory,
    VolatileAccumulativeMemory,
    VolatileMemory,
)

__all__ = [
    "auto_retry",
    "ConstantMemory",
    "traced",
    "VolatileAccumulativeMemory",
    "VolatileMemory",
    "VolatileVectorIndex",
]



from collections.abc import Callable, Coroutine
from typing import Any, Protocol, cast, final, overload
from uuid import uuid4

from draive.lmm.tools.errors import ToolError, ToolException
from draive.lmm.tools.specification import ToolSpecification
from draive.lmm.tools.status import ToolContext, ToolStatus
from draive.metrics import ArgumentsTrace, ResultTrace
from draive.parameters import ParameterSpecification, ParametrizedFunction
from draive.scope import ctx
from draive.types import MultimodalContent, MultimodalContentConvertible
from draive.utils import freeze, not_missing

__all__ = [
    "AnyTool",
    "tool",
    "Tool",
    "ToolAvailabilityCheck",
]


class ToolAvailabilityCheck(Protocol):
    def __call__(self) -> bool: ...


@final
class Tool[**Args, Result](ParametrizedFunction[Args, Coroutine[Any, Any, Result]]):
    def __init__(  # noqa: PLR0913
        self,
        /,
        name: str,
        *,
        function: Callable[Args, Coroutine[Any, Any, Result]],
        description: str | None = None,
        availability_check: ToolAvailabilityCheck | None = None,
        format_result: Callable[[Result], MultimodalContent | MultimodalContentConvertible],
        format_failure: Callable[[Exception], MultimodalContent | MultimodalContentConvertible],
        direct_result: bool = False,
    ) -> None:
        super().__init__(function=function)
        aliased_required: list[str] = []
        parameters: dict[str, ParameterSpecification] = {}
        for parameter in self._parameters.values():
            if not_missing(parameter.specification):
                parameters[parameter.aliased or parameter.name] = parameter.specification

            else:
                raise TypeError(
                    f"{function.__qualname__} can't be represented as a tool"
                    f" - argument '{parameter.name}' is missing specification."
                )

            if not (parameter.has_default or parameter.allows_missing):
                aliased_required.append(parameter.aliased or parameter.name)

        self.specification: ToolSpecification = {
            "type": "function",
            "function": {
                "name": name,
                "parameters": {
                    "type": "object",
                    "properties": parameters,
                    "required": aliased_required,
                },
                "description": description or "",
            },
        }

        self.name: str = name
        self._direct_result: bool = direct_result
        self._check_availability: ToolAvailabilityCheck = availability_check or (
            lambda: True  # available by default
        )
        self.format_result: Callable[[Result], MultimodalContent | MultimodalContentConvertible] = (
            format_result
        )
        self.format_failure: Callable[
            [Exception], MultimodalContent | MultimodalContentConvertible
        ] = format_failure

        freeze(self)

    @property
    def available(self) -> bool:
        try:
            return self._check_availability()

        except Exception:
            return False

    @property
    def requires_direct_result(self) -> bool:
        return self._direct_result

    # call from toolbox
    async def _toolbox_call(
        self,
        call_id: str,
        /,
        arguments: dict[str, Any],
        report_status: Callable[[ToolStatus], Coroutine[None, None, None]],
    ) -> MultimodalContent:
        context: ToolContext = ToolContext(
            call_id=call_id,
            tool=self.name,
            report_status=report_status,
        )
        with ctx.nested(
            self.name,
            state=[context],
            metrics=[ArgumentsTrace.of(**arguments)],
        ):
            await context.report_status(
                ToolStatus(
                    identifier=context.call_id,
                    tool=context.tool,
                    status="STARTED",
                )
            )

            try:
                if not self.available:
                    raise ToolException(f"{self.name} is not available!")

                result: Result = await super().__call__(**arguments)  # pyright: ignore[reportCallIssue]
                ctx.record(ResultTrace.of(result))

                await context.report_status(
                    ToolStatus(
                        identifier=context.call_id,
                        tool=context.tool,
                        status="FINISHED",
                    )
                )

                return MultimodalContent.of(self.format_result(result))

            except Exception as exc:
                await context.report_status(
                    ToolStatus(
                        identifier=context.call_id,
                        tool=context.tool,
                        status="FAILED",
                    )
                )
                # return an error with formatted content
                raise ToolError(
                    f"Tool {self.name}[{call_id}] failed",
                    content=MultimodalContent.of(self.format_failure(exc)),
                ) from exc

    # regular call when using as a function
    async def __call__(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        with ctx.nested(
            self.name,
            state=[
                ctx.state(  # when called as a function preserve current context
                    ToolContext,
                    default=ToolContext(
                        call_id=uuid4().hex,
                        tool=self.name,
                    ),
                )
            ],
            metrics=[ArgumentsTrace.of(*args, **kwargs)],
        ):
            result: Result = await super().__call__(
                *args,
                **kwargs,
            )

            ctx.record(ResultTrace.of(result))

            return result


AnyTool = Tool[Any, Any]


class ToolWrapper(Protocol):
    def __call__[**Args, Result](
        self,
        function: Callable[Args, Coroutine[Any, Any, Result]],
    ) -> Tool[Args, Result]: ...


class PartialToolWrapper[Result](Protocol):
    def __call__[**Args](
        self,
        function: Callable[Args, Coroutine[Any, Any, Result]],
    ) -> Tool[Args, Result]: ...


@overload
def tool[**Args, Result](
    function: Callable[Args, Coroutine[Any, Any, Result]],
    /,
) -> Tool[Args, Result]:
    """
    Convert a function to a tool using default parameters and no description.

    In order to adjust the arguments behavior and specification use an instance of Argument
    as a default value of any given argument with desired configuration
    for each argument individually.

    Parameters
    ----------
    function: Callable[Args, Coroutine[None, None, Result]]
        a function to be wrapped as a Tool.
    Returns
    -------
    Tool[Args, Result]
        a Tool representation of the provided function.
    """


@overload
def tool[Result](
    *,
    name: str | None = None,
    description: str | None = None,
    availability_check: ToolAvailabilityCheck | None = None,
    format_result: Callable[[Result], MultimodalContent | MultimodalContentConvertible],
    format_failure: Callable[[Exception], MultimodalContent | MultimodalContentConvertible]
    | None = None,
    direct_result: bool = False,
) -> PartialToolWrapper[Result]:
    """
    Convert a function to a tool using provided parameters.

    In order to adjust the arguments behavior and specification use an instance of Argument
    as a default value of any given argument with desired configuration
    for each argument individually.

    Parameters
    ----------
    name: str
        name to be used in a tool specification.
        Default is the name of the wrapped function.
    description: int
        description to be used in a tool specification. Allows to present the tool behavior to the
        external system.
        Default is empty.
    availability_check: ToolAvailabilityCheck
        function used to verify availability of the tool in given context. It can be used to check
        permissions or occurrence of a specific state to allow its usage.
        Provided function should raise an Exception when the tool should not be available.
        Default is always available.
    format_result: Callable[[Result], MultimodalContent]
        function converting tool result to MultimodalContent. It is used to format the result
        for model processing. Default implementation converts the result to string if needed.
    format_failure: Callable[[Exception], MultimodalContent]
        function converting tool call exception to a fallback MultimodalContent.
        Default implementation return "ERROR" string and logs the exception.
    direct_result: bool
        controls if tool result should break the ongoing processing and be the direct result of it.
        Note that during concurrent execution of multiple tools the call/result order defines
        direct result and exact behavior is not defined.
        Default is False.

    Returns
    -------
    Callable[[Function[Args, Coroutine[None, None, Result]]], Tool[Args, Result]]
        function allowing to convert other function to a Tool using provided configuration.
    """


@overload
def tool(
    *,
    name: str | None = None,
    description: str | None = None,
    availability_check: ToolAvailabilityCheck | None = None,
    format_failure: Callable[[Exception], MultimodalContent | MultimodalContentConvertible]
    | None = None,
    direct_result: bool = False,
) -> ToolWrapper:
    """
    Convert a function to a tool using provided parameters.

    In order to adjust the arguments behavior and specification use an instance of Argument
    as a default value of any given argument with desired configuration
    for each argument individually.

    Parameters
    ----------
    name: str
        name to be used in a tool specification.
        Default is the name of the wrapped function.
    description: int
        description to be used in a tool specification. Allows to present the tool behavior to the
        external system.
        Default is empty.
    availability_check: ToolAvailabilityCheck
        function used to verify availability of the tool in given context. It can be used to check
        permissions or occurrence of a specific state to allow its usage.
        Provided function should raise an Exception when the tool should not be available.
        Default is always available.
    format_result: Callable[[Result], MultimodalContent]
        function converting tool result to MultimodalContent. It is used to format the result
        for model processing. Default implementation converts the result to string if needed.
    format_failure: Callable[[Exception], MultimodalContent]
        function converting tool call exception to a fallback MultimodalContent.
        Default implementation return "ERROR" string and logs the exception.
    direct_result: bool
        controls if tool result should break the ongoing processing and be the direct result of it.
        Note that during concurrent execution of multiple tools the call/result order defines
        direct result and exact behavior is not defined.
        Default is False.

    Returns
    -------
    Callable[[Function[Args, Coroutine[None, None, Result]]], Tool[Args, Result]]
        function allowing to convert other function to a Tool using provided configuration.
    """


def tool[**Args, Result](  # noqa: PLR0913
    function: Callable[Args, Coroutine[Any, Any, Result]] | None = None,
    *,
    name: str | None = None,
    description: str | None = None,
    availability_check: ToolAvailabilityCheck | None = None,
    format_result: Callable[[Result], MultimodalContent | MultimodalContentConvertible]
    | None = None,
    format_failure: Callable[[Exception], MultimodalContent | MultimodalContentConvertible]
    | None = None,
    direct_result: bool = False,
) -> ToolWrapper | Tool[Args, Result]:
    def wrap(
        function: Callable[Args, Coroutine[Any, Any, Result]],
    ) -> Tool[Args, Result]:
        return Tool(
            name=name or function.__name__,
            description=description,
            function=function,
            availability_check=availability_check,
            format_result=format_result or _default_result_format,
            format_failure=format_failure or _default_failure_result,
            direct_result=direct_result,
        )

    if function := function:
        return wrap(function=function)

    else:
        return cast(PartialToolWrapper[Result], wrap)


def _default_result_format(result: Any) -> MultimodalContent:
    match result:
        case MultimodalContent() as content:
            return content

        case element if isinstance(element, MultimodalContentConvertible):
            return MultimodalContent.of(element)

        case other:
            return MultimodalContent.of(str(other))


def _default_failure_result(exception: Exception) -> MultimodalContent:
    ctx.log_error("Tool call failure", exception=exception)
    return MultimodalContent.of("ERROR")



from asyncio import AbstractEventLoop, Future, Task, TimerHandle, get_running_loop
from collections.abc import Callable, Coroutine

from draive.utils.mimic import mimic_function

__all__ = [
    "with_timeout",
]


def with_timeout[**Args, Result](
    timeout: float,
    /,
) -> Callable[
    [Callable[Args, Coroutine[None, None, Result]]],
    Callable[Args, Coroutine[None, None, Result]],
]:
    """\
    Simple timeout wrapper for the function call. \
    When the timeout time will pass before function returns it will be \
    cancelled and TimeoutError exception will raise. Make sure that wrapped \
    function handles cancellation properly.
    This wrapper is not thread safe.

    Parameters
    ----------
    timeout: float
        timeout time in seconds

    Returns
    -------
    Callable[[Callable[_Args, _Result]], Callable[_Args, _Result]] | Callable[_Args, _Result]
        function wrapper adding timeout
    """

    def _wrap(
        function: Callable[Args, Coroutine[None, None, Result]],
    ) -> Callable[Args, Coroutine[None, None, Result]]:
        return _AsyncTimeout(
            function,
            timeout=timeout,
        )

    return _wrap


class _AsyncTimeout[**Args, Result]:
    def __init__(
        self,
        function: Callable[Args, Coroutine[None, None, Result]],
        /,
        timeout: float,
    ) -> None:
        self._function: Callable[Args, Coroutine[None, None, Result]] = function
        self._timeout: float = timeout

        # mimic function attributes if able
        mimic_function(function, within=self)

    async def __call__(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        loop: AbstractEventLoop = get_running_loop()
        future: Future[Result] = loop.create_future()
        task: Task[Result] = loop.create_task(
            self._function(
                *args,
                **kwargs,
            ),
        )

        def on_timeout(
            future: Future[Result],
        ) -> None:
            if future.done():
                return  # ignore if already finished

            # result future on its completion will ensure that task will complete
            future.set_exception(TimeoutError())

        timeout_handle: TimerHandle = loop.call_later(
            self._timeout,
            on_timeout,
            future,
        )

        def on_completion(
            task: Task[Result],
        ) -> None:
            timeout_handle.cancel()  # at this stage we no longer need timeout to trigger

            if future.done():
                return  # ignore if already finished

            try:
                future.set_result(task.result())

            except Exception as exc:
                future.set_exception(exc)

        task.add_done_callback(on_completion)

        def on_result(
            future: Future[Result],
        ) -> None:
            task.cancel()  # when result future completes make sure that task also completes

        future.add_done_callback(on_result)

        return await future



from collections.abc import Callable, Sequence

from draive.evaluation import EvaluationScore, evaluator

__all__ = [
    "text_keywords_evaluator",
]


@evaluator(name="text_keywords")
async def text_keywords_evaluator(
    text: str,
    /,
    keywords: Sequence[str],
    normalization: Callable[[str], str] | None = None,
) -> EvaluationScore:
    if not text:
        return EvaluationScore(
            value=0,
            comment="Input text was empty!",
        )

    if not keywords:
        return EvaluationScore(
            value=0,
            comment="Keywords list was empty!",
        )

    text_normalization: Callable[[str], str]
    if normalization is not None:
        text_normalization = normalization

    else:
        text_normalization = _lowercased

    normalized_text: str = text_normalization(text)
    return EvaluationScore(
        value=len(
            [keyword for keyword in keywords if text_normalization(keyword) in normalized_text]
        )
        / len(keywords),
    )


def _lowercased(text: str) -> str:
    return text.lower()



from collections.abc import AsyncGenerator, Sequence
from copy import copy
from typing import Any, Literal, cast, overload
from uuid import uuid4

from draive.gemini.client import GeminiClient
from draive.gemini.config import GeminiConfig
from draive.gemini.errors import GeminiException
from draive.gemini.models import (
    GeminiChoice,
    GeminiDataMessageContent,
    GeminiDataReferenceMessageContent,
    GeminiFunctionCallMessageContent,
    GeminiFunctionsTool,
    GeminiFunctionToolSpecification,
    GeminiGenerationResult,
    GeminiMessage,
    GeminiMessageContent,
    GeminiRequestMessage,
    GeminiTextMessageContent,
)
from draive.instructions import Instruction
from draive.lmm import LMMToolSelection, ToolSpecification
from draive.metrics import ArgumentsTrace, ResultTrace, TokenUsage
from draive.parameters import DataModel
from draive.scope import ctx
from draive.types import (
    AudioBase64Content,
    AudioURLContent,
    ImageBase64Content,
    ImageURLContent,
    LMMCompletion,
    LMMCompletionChunk,
    LMMContextElement,
    LMMInput,
    LMMOutput,
    LMMOutputStream,
    LMMOutputStreamChunk,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponse,
    MultimodalContentElement,
    TextContent,
    VideoBase64Content,
    VideoURLContent,
)

__all__ = [
    "gemini_lmm_invocation",
]


@overload
async def gemini_lmm_invocation(
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: Literal[True],
    **extra: Any,
) -> LMMOutputStream: ...


@overload
async def gemini_lmm_invocation(
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: Literal[False] = False,
    **extra: Any,
) -> LMMOutput: ...


@overload
async def gemini_lmm_invocation(
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: bool = False,
    **extra: Any,
) -> LMMOutputStream | LMMOutput: ...


async def gemini_lmm_invocation(  # noqa: PLR0913
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: bool = False,
    **extra: Any,
) -> LMMOutputStream | LMMOutput:
    with ctx.nested(
        "gemini_lmm_invocation",
        metrics=[
            ArgumentsTrace.of(
                instruction=instruction,
                context=context,
                tools=tools,
                tool_selection=tool_selection,
                output=output,
                stream=stream,
                **extra,
            ),
        ],
    ):
        ctx.log_debug("Requested Gemini lmm")
        client: GeminiClient = ctx.dependency(GeminiClient)
        config: GeminiConfig = ctx.state(GeminiConfig).updated(**extra)
        ctx.record(config)

        match output:
            case "text":
                config = config.updated(response_format="text/plain")

            case "json":
                if tools:
                    ctx.log_warning(
                        "Attempting to use Gemini in JSON mode with tools which is not supported."
                        " Using text mode instead..."
                    )
                    config = config.updated(response_format="text/plain")

                else:
                    config = config.updated(response_format="application/json")

        messages: list[GeminiRequestMessage] = [
            _convert_context_element(element=element) for element in context
        ]

        if stream:
            return ctx.stream(
                _generation_stream(
                    client=client,
                    config=config,
                    instruction=Instruction.of(instruction).format(),
                    messages=messages,
                    tools=tools,
                    tool_selection=tool_selection,
                ),
            )

        else:
            return await _generate(
                client=client,
                config=config,
                instruction=Instruction.of(instruction).format(),
                messages=messages,
                tools=tools,
                tool_selection=tool_selection,
            )


def _convert_content_element(  # noqa: PLR0911
    element: MultimodalContentElement,
) -> dict[str, Any]:
    match element:
        case TextContent() as text:
            return {"text": text.text}

        case ImageURLContent() as image:
            return {
                "fileData": {
                    "mimeType": image.mime_type or "image",
                    "fileUri": image.image_url,
                }
            }

        case ImageBase64Content() as image:
            return {
                "inlineData": {
                    "mimeType": image.mime_type or "image",
                    "data": image.image_base64,
                }
            }

        case AudioURLContent() as audio:
            return {
                "fileData": {
                    "mimeType": audio.mime_type or "audio",
                    "fileUri": audio.audio_url,
                }
            }

        case AudioBase64Content() as audio:
            return {
                "inlineData": {
                    "mimeType": audio.mime_type or "audio",
                    "data": audio.audio_base64,
                }
            }

        case VideoURLContent() as video:
            return {
                "fileData": {
                    "mimeType": video.mime_type or "video",
                    "fileUri": video.video_url,
                }
            }

        case VideoBase64Content() as video:
            return {
                "inlineData": {
                    "mimeType": video.mime_type or "video",
                    "data": video.video_base64,
                }
            }

        case DataModel() as data:
            return {"text": data.as_json()}


def _convert_context_element(
    element: LMMContextElement,
) -> GeminiRequestMessage:
    match element:
        case LMMInput() as input:
            return {
                "role": "user",
                "parts": [
                    _convert_content_element(element=element) for element in input.content.parts
                ],
            }

        case LMMCompletion() as completion:
            return {
                "role": "model",
                "parts": [
                    _convert_content_element(element=element)
                    for element in completion.content.parts
                ],
            }

        case LMMToolRequests() as tool_requests:
            return {
                "role": "model",
                "parts": [
                    {
                        "functionCall": {
                            "name": request.tool,
                            "args": request.arguments,
                        },
                    }
                    for request in tool_requests.requests
                ],
            }

        case LMMToolResponse() as tool_response:
            return {
                "role": "model",
                "parts": [
                    {
                        "functionResponse": {
                            "name": tool_response.tool,
                            "response": tool_response.content.as_dict(),
                        },
                    }
                ],
            }


def _convert_content_part(  # noqa: PLR0911
    part: GeminiMessageContent,
) -> MultimodalContentElement:
    match part:
        case GeminiTextMessageContent() as text:
            return TextContent(text=text.text)

        case GeminiDataMessageContent() as data:
            mime_type: str = data.data.mime_type
            if mime_type.startswith("image"):
                return ImageBase64Content(
                    mime_type=cast(
                        Literal["image/jpeg", "image/png", "image/gif"] | None,
                        mime_type
                        if mime_type in {"image/jpeg", "image/png", "image/gif"}
                        else None,
                    ),
                    image_base64=data.data.data,
                )

            elif mime_type.startswith("audio"):
                return AudioBase64Content(
                    mime_type=mime_type,
                    audio_base64=data.data.data,
                )

            elif mime_type.startswith("video"):
                return VideoBase64Content(
                    mime_type=mime_type,
                    video_base64=data.data.data,
                )

            else:
                raise GeminiException("Unsupported result content data %s", data)

        case GeminiDataReferenceMessageContent() as reference:
            mime_type: str = reference.reference.mime_type
            if mime_type.startswith("image"):
                return ImageURLContent(
                    mime_type=cast(
                        Literal["image/jpeg", "image/png", "image/gif"] | None,
                        mime_type
                        if mime_type in {"image/jpeg", "image/png", "image/gif"}
                        else None,
                    ),
                    image_url=reference.reference.uri,
                )

            elif mime_type.startswith("audio"):
                return AudioURLContent(
                    mime_type=mime_type,
                    audio_url=reference.reference.uri,
                )

            elif mime_type.startswith("video"):
                return VideoURLContent(
                    mime_type=mime_type,
                    video_url=reference.reference.uri,
                )

            else:
                raise GeminiException("Unsupported result content reference %s", reference)

        case other:
            raise GeminiException("Unsupported result content %s", other)


async def _generate(  # noqa: PLR0913, C901, PLR0912, PLR0915
    *,
    client: GeminiClient,
    config: GeminiConfig,
    instruction: str,
    messages: list[GeminiRequestMessage],
    tools: Sequence[ToolSpecification] | None,
    tool_selection: LMMToolSelection,
) -> LMMOutput:
    result: GeminiGenerationResult
    converted_tools: Sequence[GeminiFunctionToolSpecification] = []
    for tool in tools or []:
        tool_function: GeminiFunctionToolSpecification = cast(
            # those models are the same, can safely cast
            GeminiFunctionToolSpecification,
            tool["function"],
        )
        # AIStudio api requires to delete properties if those are empty...
        if "parameters" in tool_function and not tool_function["parameters"]["properties"]:
            tool_function = copy(tool_function)
            del tool_function["parameters"]

        converted_tools.append(tool_function)

    match tool_selection:
        case "auto":
            result = await client.generate(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[GeminiFunctionsTool(functionDeclarations=converted_tools)],
                tool_calling_mode="AUTO",
            )

        case "none":
            result = await client.generate(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[],
                tool_calling_mode="NONE",
            )

        case "required":
            result = await client.generate(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[GeminiFunctionsTool(functionDeclarations=converted_tools)],
                tool_calling_mode="ANY",
            )

        case tool:
            assert tool in (tools or []), "Can't suggest a tool without using it"  # nosec: B101
            tool_function: GeminiFunctionToolSpecification = cast(
                # those models are the same, can safely cast
                GeminiFunctionToolSpecification,
                tool["function"],
            )
            # AIStudio api requires to delete properties if those are empty...
            if "parameters" in tool_function and not tool_function["parameters"]["properties"]:
                tool_function = copy(tool_function)
                del tool_function["parameters"]

            result = await client.generate(
                config=config,
                instruction=instruction,
                messages=messages,
                tools=[GeminiFunctionsTool(functionDeclarations=[tool_function])],
                tool_calling_mode="ANY",
            )

    if usage := result.usage:
        ctx.record(
            TokenUsage.for_model(
                config.model,
                input_tokens=usage.prompt_tokens,
                output_tokens=usage.generated_tokens,
            ),
        )

    if not result.choices:
        raise GeminiException("Invalid Gemini completion - missing messages!", result)

    result_choice: GeminiChoice = result.choices[0]
    result_message: GeminiMessage
    match result_choice.finish_reason:
        case "STOP":
            if message := result_choice.content:
                result_message = message

            else:
                raise GeminiException(
                    "Invalid Gemini response - missing message content: %s", result
                )

        case "MAX_TOKENS":
            raise GeminiException("Gemini response finish caused by token limit: %s", result)

        case "SAFETY":
            raise GeminiException("Gemini response finish caused by safety reason: %s", result)

        case "RECITATION":
            raise GeminiException("Gemini response finish caused by recitation reason: %s", result)

        case "OTHER":
            raise GeminiException("Gemini response finish caused by unknown reason: %s", result)

    message_parts: list[
        GeminiTextMessageContent | GeminiDataReferenceMessageContent | GeminiDataMessageContent
    ] = []
    tool_calls: list[GeminiFunctionCallMessageContent] = []
    for part in result_message.content:
        match part:
            case GeminiTextMessageContent() as text:
                message_parts.append(text)

            case GeminiFunctionCallMessageContent() as call:
                tool_calls.append(call)

            case GeminiDataReferenceMessageContent() as reference:
                message_parts.append(reference)

            case GeminiDataMessageContent() as data:
                message_parts.append(data)

            case other:
                raise GeminiException("Invalid Gemini completion part", other)

    if tool_calls and (tools := tools):
        if __debug__ and message_parts:
            ctx.log_debug("Gemini has generated a message and tool calls, ignoring the message...")

        ctx.record(ResultTrace.of(tool_calls))

        return LMMToolRequests(
            requests=[
                LMMToolRequest(
                    identifier=uuid4().hex,
                    tool=call.function_call.name,
                    arguments=call.function_call.arguments,
                )
                for call in tool_calls
            ]
        )

    elif message_parts:
        ctx.record(ResultTrace.of(message_parts))
        return LMMCompletion.of(*[_convert_content_part(part) for part in message_parts])

    else:
        raise GeminiException("Invalid Gemini completion", result)


async def _generation_stream(  # noqa: PLR0913
    *,
    client: GeminiClient,
    config: GeminiConfig,
    instruction: str,
    messages: list[GeminiRequestMessage],
    tools: Sequence[ToolSpecification] | None,
    tool_selection: LMMToolSelection,
) -> AsyncGenerator[LMMOutputStreamChunk, None]:
    ctx.log_debug("Gemini streaming api is not supported yet, using regular response...")
    output: LMMOutput = await _generate(
        client=client,
        config=config,
        instruction=instruction,
        messages=messages,
        tools=tools,
        tool_selection=tool_selection,
    )

    match output:
        case LMMCompletion() as completion:
            yield LMMCompletionChunk.of(completion.content)

        case other:
            yield other



from draive import markdown_block, markdown_blocks


def test_returns_none_with_empty():
    source: str = ""

    assert markdown_block(source=source) is None


def test_info_returns_none_with_empty():
    source: str = ""

    assert markdown_block("test", source=source) is None


def test_returns_none_without_block():
    source: str = "Lorem ipsum"

    assert markdown_block(source=source) is None


def test_info_returns_none_without_block():
    source: str = "Lorem ipsum"

    assert markdown_block("test", source=source) is None


def test_returns_none_with_one_line_block():
    source: str = "`Lorem ipsum`"

    assert markdown_block(source=source) is None


def test_info_returns_none_with_one_line_block():
    source: str = "`test Lorem ipsum`"

    assert markdown_block("test", source=source) is None


def test_returns_content_with_no_info_block():
    source: str = "``` Lorem ipsum```"

    assert markdown_block(source=source) == "Lorem ipsum"


def test_returns_content_with_any_info_block():
    source: str = "```other Lorem ipsum```"

    assert markdown_block(source=source) == "Lorem ipsum"


def test_info_returns_none_with_no_info_block():
    source: str = "``` Lorem ipsum```"

    assert markdown_block("test", source=source) is None


def test_info_returns_none_with_other_info_block():
    source: str = "```other Lorem ipsum```"

    assert markdown_block("test", source=source) is None


def test_returns_none_with_block_closing():
    source: str = "Lorem ipsum```"

    assert markdown_block(source=source) is None


def test_info_returns_none_with_block_closing():
    source: str = "Lorem ipsum```"

    assert markdown_block("test", source=source) is None


def test_info_returns_none_with_missing_info():
    source: str = "```Lorem ipsum```"

    assert markdown_block("test", source=source) is None


def test_returns_none_without_block_closing():
    source: str = "``` Lorem ipsum"

    assert markdown_block(source=source) is None


def test_info_returns_none_without_block_closing():
    source: str = "```test Lorem ipsum"

    assert markdown_block("test", source=source) is None


def test_returns_none_with_malformed_block_opening():
    source: str = "`` Lorem ipsum```"

    assert markdown_block("test", source=source) is None


def test_info_returns_none_with_malformed_block_opening():
    source: str = "``test Lorem ipsum```"

    assert markdown_block("test", source=source) is None


def test_returns_none_with_malformed_block_closing():
    source: str = "```test Lorem ipsum``"

    assert markdown_block(source=source) is None


def test_info_returns_none_with_malformed_block_closing():
    source: str = "```test Lorem ipsum``"

    assert markdown_block("test", source=source) is None


def test_returns_content_with_valid_block():
    source: str = "``` Lorem ipsum```"

    assert markdown_block(source=source) == "Lorem ipsum"


def test_info_returns_content_with_valid_block():
    source: str = "```test Lorem ipsum```"

    assert markdown_block("test", source=source) == "Lorem ipsum"


def test_returns_content_with_surrounded_block():
    source: str = "Lorem``` Lorem ipsum``` ipsum"

    assert markdown_block(source=source) == "Lorem ipsum"


def test_info_returns_content_with_surrounded_block():
    source: str = "Lorem```test Lorem ipsum``` ipsum"

    assert markdown_block("test", source=source) == "Lorem ipsum"


def test_returns_first_content_with_multiple_blocks():
    source: str = "``` Lorem ipsum``` ``` Other```"

    assert markdown_block(source=source) == "Lorem ipsum"


def test_info_returns_first_content_with_multiple_blocks():
    source: str = "```test Lorem ipsum``` ```test Other```"

    assert markdown_block("test", source=source) == "Lorem ipsum"


def test_returns_outer_content_with_nested_blocks():
    source: str = "```outer Other ```inner Lorem ipsum``` ```"

    assert markdown_block(source=source) == "Other ```inner Lorem ipsum"


def test_info_returns_outer_content_with_nested_blocks():
    source: str = "```test Other ```test Lorem ipsum``` ```"

    assert markdown_block("test", source=source) == "Other ```test Lorem ipsum"


def test_returns_content_with_extended_block_closing():
    source: str = "``` Lorem`````"

    assert markdown_block(source=source) == "Lorem``"


def test_info_returns_content_with_extended_block_closing():
    source: str = "```test Lorem`````"

    assert markdown_block("test", source=source) == "Lorem``"


def test_returns_content_with_multiple_blocks():
    source: str = "``` Lorem ipsum``` ```test Dolor``` ```more Sit amet```"
    contents: list[str] = list(markdown_blocks(source=source))

    assert contents == ["Lorem ipsum", "Dolor", "Sit amet"]


def test_info_returns_content_with_multiple_blocks():
    source: str = "```test Lorem ipsum``` ```test Dolor``` ```test Sit amet```"
    contents: list[str] = list(markdown_blocks("test", source=source))

    assert contents == ["Lorem ipsum", "Dolor", "Sit amet"]


def test_info_returns_content_with_multiple_blocks_skipping_other_blocks():
    source: str = "```test Lorem ipsum``` ```other Dolor``` ```test Sit amet```"
    contents: list[str] = list(markdown_blocks("test", source=source))

    assert contents == ["Lorem ipsum", "Sit amet"]



from draive.generation.text.generator import TextGenerator
from draive.generation.text.lmm import lmm_generate_text
from draive.parameters import State

__all__ = [
    "TextGeneration",
]


class TextGeneration(State):
    generate: TextGenerator = lmm_generate_text



from draive.embedding import Embedded, embed_texts
from draive.evaluation import EvaluationScore, evaluator
from draive.evaluators.score import CommonScoreModel
from draive.generation import generate_model
from draive.similarity.score import vector_similarity_score

__all__ = [
    "text_similarity_evaluator",
    "text_vector_similarity_evaluator",
]


INSTRUCTION: str = """\
You will be given two texts: a reference text and a compared text. \
Your task is to rate the compared text using only the Similarity metric, \
which is described in the Evaluation Criteria.
Please make sure you read and understand these instructions very carefully.
Keep this document open while reviewing, and refer to it as needed.

Evaluation Criteria:
Similarity (0.0-2.0) - the degree of semantic similarity between the reference text \
and the compared text.

Rating Scale:
0.0: No similarity - the reference text and compared text are completely unrelated in meaning.
1.0: Moderate similarity - the reference text and compared text share some common themes or ideas.
2.0: High similarity - the reference text and compared text are very close in meaning \
or convey the same information.

Evaluation Steps:
1. Read both the reference text and the compared text carefully.
2. Compare the semantic meaning of the reference text and the compared text.
3. Assign a similarity score from 0.0 to 2.0 based on the provided criteria.

Important: The score must be a decimal number from 0.0 to 2.0. 2.0 is the maximum, \
do not exceed this value.
"""


INPUT_TEMPLATE: str = """
<REFERENCE_TEXT>
{reference}
</REFERENCE_TEXT>

<COMPARED_TEXT>
{compared}
</COMPARED_TEXT>
"""


@evaluator(name="text_similarity")
async def text_similarity_evaluator(
    compared: str,
    /,
    reference: str,
) -> EvaluationScore:
    if not compared:
        return EvaluationScore(
            value=0,
            comment="Input text was empty!",
        )

    if not reference:
        return EvaluationScore(
            value=0,
            comment="Reference text was empty!",
        )

    score: CommonScoreModel = await generate_model(
        CommonScoreModel,
        instruction=INSTRUCTION,
        input=INPUT_TEMPLATE.format(
            reference=reference,
            compared=compared,
        ),
        examples=[
            (
                INPUT_TEMPLATE.format(
                    reference=(
                        "Cats are popular pets. They are independent and like to groom themselves."
                    ),
                    compared=(
                        "Bananas are a healthy fruit. They are rich in potassium and easy to peel."
                    ),
                ),
                CommonScoreModel(score=0.0),
            ),
            (
                INPUT_TEMPLATE.format(
                    reference=(
                        "The beach is a great place for relaxation. "
                        "People enjoy swimming and sunbathing."
                    ),
                    compared=(
                        "Many people like to spend time outdoors. "
                        "Parks are popular for picnics and walking."
                    ),
                ),
                CommonScoreModel(score=1.0),
            ),
            (
                INPUT_TEMPLATE.format(
                    reference=(
                        "Coffee is a popular morning drink. It contains caffeine which helps "
                        "people feel more alert."
                    ),
                    compared=(
                        "Many people start their day with coffee. "
                        "The caffeine in coffee can increase alertness and energy."
                    ),
                ),
                CommonScoreModel(score=2.0),
            ),
        ],
    )

    return score.normalized(divider=2)


@evaluator(name="text_vector_similarity")
async def text_vector_similarity_evaluator(
    compared: str,
    /,
    reference: str,
) -> float:
    embedding: list[Embedded[str]] = await embed_texts([reference, compared])

    return vector_similarity_score(embedding[0].vector, embedding[1].vector)



from collections.abc import Callable
from typing import Any

import numpy as np
from numpy.typing import NDArray

from draive.similarity.cosine import cosine_similarity

__all__ = [
    "vector_similarity_search",
]


def vector_similarity_search(
    query_vector: NDArray[Any] | list[float],
    values_vectors: list[NDArray[Any]] | list[list[float]],
    limit: int,
    score_threshold: float | None = None,
    similarity: Callable[
        [list[NDArray[Any]] | NDArray[Any], list[NDArray[Any]] | NDArray[Any]], NDArray[Any]
    ] = cosine_similarity,
) -> list[int]:
    assert limit > 0  # nosec: B101
    if not values_vectors:
        return []

    query: NDArray[Any] = np.array(query_vector)
    if query.ndim == 1:
        query = np.expand_dims(query_vector, axis=0)
    values: NDArray[Any] = np.array(values_vectors)
    matching_scores: NDArray[Any] = similarity(values, query)
    sorted_indices: list[int] = list(reversed(np.argsort(matching_scores)))

    if score_threshold:
        return [
            int(idx)
            for idx in sorted_indices  # pyright: ignore[reportUnknownVariableType]
            if matching_scores[idx] > score_threshold  # pyright: ignore[reportUnknownArgumentType]
        ][:limit]

    else:
        return [
            int(idx)
            for idx in sorted_indices  # pyright: ignore[reportUnknownVariableType]
        ][:limit]



from draive.openai.client import OpenAIClient
from draive.openai.config import (
    OpenAIChatConfig,
    OpenAIEmbeddingConfig,
    OpenAIImageGenerationConfig,
)
from draive.openai.embedding import openai_embed_text
from draive.openai.errors import OpenAIException
from draive.openai.images import openai_generate_image
from draive.openai.lmm import openai_lmm_invocation
from draive.openai.tokenization import openai_tokenize_text

__all__ = [
    "openai_embed_text",
    "openai_generate_image",
    "openai_lmm_invocation",
    "openai_tokenize_text",
    "OpenAIChatConfig",
    "OpenAIClient",
    "OpenAIEmbeddingConfig",
    "OpenAIException",
    "OpenAIImageGenerationConfig",
]



from collections.abc import Callable
from typing import Protocol, Self, cast, final, overload, runtime_checkable

from draive.evaluation.score import Evaluation, EvaluationScore
from draive.parameters import DataModel, Field, ParameterPath
from draive.utils import freeze

__all__ = [
    "evaluator",
    "Evaluator",
    "EvaluatorResult",
    "PreparedEvaluator",
]


class EvaluatorResult(DataModel):
    evaluator: str = Field(
        description="Name of evaluator",
    )
    score: EvaluationScore = Field(
        description="Evaluation score",
    )
    threshold: float = Field(
        description="Score threshold required to pass evaluation",
    )

    @property
    def passed(self) -> bool:
        return self.score.value >= self.threshold


@runtime_checkable
class PreparedEvaluator[Value](Protocol):
    async def __call__(
        self,
        value: Value,
        /,
    ) -> EvaluatorResult: ...


@final
class Evaluator[Value, **Args]:
    def __init__(
        self,
        name: str,
        evaluation: Evaluation[Value, Args],
        threshold: float | None,
    ) -> None:
        assert (  # nosec: B101
            threshold is None or 0 <= threshold <= 1
        ), "Evaluation threshold has to be between 0 and 1"

        self._evaluation: Evaluation[Value, Args] = evaluation
        self.name: str = name
        self.threshold: float = threshold or 1

        freeze(self)

    def with_threshold(
        self,
        threshold: float,
    ) -> Self:
        return self.__class__(
            name=self.name,
            evaluation=self._evaluation,
            threshold=threshold,
        )

    def prepared(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> PreparedEvaluator[Value]:
        async def evaluate(
            value: Value,
        ) -> EvaluatorResult:
            return await self(
                value,
                *args,
                **kwargs,
            )

        return evaluate

    def contra_map[Mapped](
        self,
        mapping: Callable[[Mapped], Value] | ParameterPath[Mapped, Value] | Value,
        /,
    ) -> "Evaluator[Mapped, Args]":
        mapper: Callable[[Mapped], Value]
        match mapping:
            case Callable() as function:  # pyright: ignore[reportUnknownVariableType]
                mapper = function

            case path:
                assert isinstance(  # nosec: B101
                    path, ParameterPath
                ), "Prepare parameter path by using Self._.path.to.property"
                mapper = cast(ParameterPath[Mapped, Value], path).__call__

        async def evaluation(
            value: Mapped,
            *args: Args.args,
            **kwargs: Args.kwargs,
        ) -> EvaluationScore | float | bool:
            return await self._evaluation(
                mapper(value),
                *args,
                **kwargs,
            )

        return Evaluator[Mapped, Args](
            name=self.name,
            evaluation=evaluation,
            threshold=self.threshold,
        )

    async def __call__(
        self,
        value: Value,
        /,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> EvaluatorResult:
        evaluation_score: EvaluationScore
        match await self._evaluation(
            value,
            *args,
            **kwargs,
        ):
            case float() as score_value:
                evaluation_score = EvaluationScore(value=score_value)

            case bool() as score_bool:
                evaluation_score = EvaluationScore(value=1 if score_bool else 0)

            case EvaluationScore() as score:
                evaluation_score = score

            # for whatever reason pyright wants int to be handled...
            case int() as score_int:
                evaluation_score = EvaluationScore(value=float(score_int))

        return EvaluatorResult(
            evaluator=self.name,
            score=evaluation_score,
            threshold=self.threshold,
        )


@overload
def evaluator[Value, **Args](
    evaluation: Evaluation[Value, Args] | None = None,
    /,
) -> Evaluator[Value, Args]: ...


@overload
def evaluator[Value, **Args](
    *,
    name: str | None = None,
    threshold: float | None = None,
) -> Callable[
    [Evaluation[Value, Args]],
    Evaluator[Value, Args],
]: ...


def evaluator[Value, **Args](
    evaluation: Evaluation[Value, Args] | None = None,
    *,
    name: str | None = None,
    threshold: float | None = None,
) -> (
    Callable[
        [Evaluation[Value, Args]],
        Evaluator[Value, Args],
    ]
    | Evaluator[Value, Args]
):
    def wrap(
        evaluation: Evaluation[Value, Args],
    ) -> Evaluator[Value, Args]:
        return Evaluator(
            name=name or evaluation.__name__,
            evaluation=evaluation,
            threshold=threshold,
        )

    if evaluation:
        return wrap(evaluation)

    else:
        return wrap



from collections.abc import AsyncGenerator, Sequence
from typing import Any, Literal, overload

from draive.instructions import Instruction
from draive.lmm import LMMToolSelection, ToolSpecification
from draive.metrics import ArgumentsTrace, ResultTrace
from draive.metrics.tokens import TokenUsage
from draive.ollama.client import OllamaClient
from draive.ollama.config import OllamaChatConfig
from draive.ollama.errors import OllamaException
from draive.ollama.models import ChatCompletionResponse, ChatMessage
from draive.scope import ctx
from draive.types import (
    LMMCompletion,
    LMMCompletionChunk,
    LMMContextElement,
    LMMInput,
    LMMOutput,
    LMMOutputStream,
    LMMOutputStreamChunk,
    LMMToolRequests,
    LMMToolResponse,
)

__all__ = [
    "ollama_lmm_invocation",
]


@overload
async def ollama_lmm_invocation(
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: Literal[True],
    **extra: Any,
) -> LMMOutputStream: ...


@overload
async def ollama_lmm_invocation(
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: Literal[False] = False,
    **extra: Any,
) -> LMMOutput: ...


@overload
async def ollama_lmm_invocation(
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: bool = False,
    **extra: Any,
) -> LMMOutputStream | LMMOutput: ...


async def ollama_lmm_invocation(  # noqa: PLR0913
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: bool = False,
    **extra: Any,
) -> LMMOutputStream | LMMOutput:
    with ctx.nested(
        "ollama_lmm_invocation",
        metrics=[
            ArgumentsTrace.of(
                instruction=instruction,
                context=context,
                tools=tools,
                tool_selection=tool_selection,
                output=output,
                stream=stream,
                **extra,
            ),
        ],
    ):
        ctx.log_debug("Requested Ollama lmm")
        client: OllamaClient = ctx.dependency(OllamaClient)
        config: OllamaChatConfig = ctx.state(OllamaChatConfig).updated(**extra)
        ctx.record(config)

        if tools:
            ctx.log_warning(
                "Attempting to use Ollama with tools which is not supported."
                " Ignoring provided tools..."
            )

        match output:
            case "text":
                config = config.updated(response_format="text")

            case "json":
                config = config.updated(response_format="json")

        messages: list[ChatMessage] = [
            ChatMessage(
                role="system",
                content=Instruction.of(instruction).format(),
            ),
            *[_convert_context_element(element=element) for element in context],
        ]

        if stream:
            return ctx.stream(
                _chat_completion_stream(
                    client=client,
                    config=config,
                    messages=messages,
                ),
            )

        else:
            return await _chat_completion(
                client=client,
                config=config,
                messages=messages,
            )


def _convert_context_element(
    element: LMMContextElement,
) -> ChatMessage:
    match element:
        case LMMInput() as input:
            return ChatMessage(
                role="user",
                content=input.content.as_string(),
            )

        case LMMCompletion() as completion:
            return ChatMessage(
                role="assistant",
                content=completion.content.as_string(),
            )

        case LMMToolRequests():
            raise NotImplementedError("Tools use is not supported by Ollama")

        case LMMToolResponse():
            raise NotImplementedError("Tools use is not supported by Ollama")


async def _chat_completion(
    *,
    client: OllamaClient,
    config: OllamaChatConfig,
    messages: list[ChatMessage],
) -> LMMOutput:
    completion: ChatCompletionResponse = await client.chat_completion(
        config=config,
        messages=messages,
    )

    ctx.record(
        TokenUsage.for_model(
            config.model,
            input_tokens=completion.prompt_eval_count,
            output_tokens=completion.eval_count,
        ),
    )

    if message := completion.message.content:
        ctx.record(ResultTrace.of(message))
        return LMMCompletion.of(message)

    else:
        raise OllamaException("Invalid Ollama completion", completion)


async def _chat_completion_stream(
    *,
    client: OllamaClient,
    config: OllamaChatConfig,
    messages: list[ChatMessage],
) -> AsyncGenerator[LMMOutputStreamChunk, None]:
    ctx.log_debug("Ollama streaming api is not supported yet, using regular response...")
    output: LMMOutput = await _chat_completion(
        client=client,
        config=config,
        messages=messages,
    )

    match output:
        case LMMCompletion() as completion:
            yield LMMCompletionChunk.of(completion.content)

        case other:
            yield other



from asyncio import AbstractEventLoop, Task, get_running_loop, iscoroutinefunction, shield
from collections import OrderedDict
from collections.abc import Callable, Coroutine, Hashable
from functools import _make_key, partial  # pyright: ignore[reportPrivateUsage]
from time import monotonic
from typing import NamedTuple, cast, overload
from weakref import ref

from draive.utils.mimic import mimic_function

__all__ = [
    "cache",
]


@overload
def cache[**Args, Result](
    function: Callable[Args, Result],
    /,
) -> Callable[Args, Result]: ...


@overload
def cache[**Args, Result](
    *,
    limit: int = 1,
    expiration: float | None = None,
) -> Callable[[Callable[Args, Result]], Callable[Args, Result]]: ...


def cache[**Args, Result](
    function: Callable[Args, Result] | None = None,
    *,
    limit: int = 1,
    expiration: float | None = None,
) -> Callable[[Callable[Args, Result]], Callable[Args, Result]] | Callable[Args, Result]:
    """\
    Simple lru function result cache with optional expire time. \
    Works for both sync and async functions. \
    It is not allowed to be used on class methods. \
    This wrapper is not thread safe.

    Parameters
    ----------
    function: Callable[_Args, _Result]
        function to wrap in cache, either sync or async
    limit: int
        limit of cache entries to keep, default is 1
    expiration: float | None
        entries expiration time in seconds, default is None (not expiring)

    Returns
    -------
    Callable[[Callable[_Args, _Result]], Callable[_Args, _Result]] | Callable[_Args, _Result]
        provided function wrapped in cache
    """

    def _wrap(function: Callable[Args, Result]) -> Callable[Args, Result]:
        if iscoroutinefunction(function):
            return cast(
                Callable[Args, Result],
                _AsyncCache(
                    function,
                    limit=limit,
                    expiration=expiration,
                ),
            )
        else:
            return cast(
                Callable[Args, Result],
                _SyncCache(
                    function,
                    limit=limit,
                    expiration=expiration,
                ),
            )

    if function := function:
        return _wrap(function)

    else:
        return _wrap


class _CacheEntry[Entry](NamedTuple):
    value: Entry
    expire: float | None


class _SyncCache[**Args, Result]:
    def __init__(
        self,
        function: Callable[Args, Result],
        /,
        limit: int,
        expiration: float | None,
    ) -> None:
        self._function: Callable[Args, Result] = function
        self._cached: OrderedDict[Hashable, _CacheEntry[Result]] = OrderedDict()
        self._limit: int = limit
        if expiration := expiration:

            def next_expire_time() -> float | None:
                return monotonic() + expiration
        else:

            def next_expire_time() -> float | None:
                return None

        self._next_expire_time: Callable[[], float | None] = next_expire_time

        # mimic function attributes if able
        mimic_function(function, within=self)

    def __get__(
        self,
        instance: object | None,
        owner: type | None = None,
        /,
    ) -> Callable[Args, Result]:
        if owner is None or instance is None:
            return self

        else:
            return mimic_function(
                self._function,
                within=partial(
                    self.__method_call__,
                    instance,
                ),
            )

    def __call__(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        key: Hashable = _make_key(
            args=args,
            kwds=kwargs,
            typed=True,
        )

        match self._cached.get(key):
            case None:
                pass

            case entry:
                if (expire := entry[1]) and expire < monotonic():
                    # if still running let it complete if able
                    del self._cached[key]  # continue the same way as if empty
                else:
                    self._cached.move_to_end(key)
                    return entry[0]

        result: Result = self._function(*args, **kwargs)
        self._cached[key] = _CacheEntry(
            value=result,
            expire=self._next_expire_time(),
        )
        if len(self._cached) > self._limit:
            # if still running let it complete if able
            self._cached.popitem(last=False)

        return result

    def __method_call__(
        self,
        __method_self: object,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        key: Hashable = _make_key(
            args=(ref(__method_self), *args),
            kwds=kwargs,
            typed=True,
        )

        match self._cached.get(key):
            case None:
                pass

            case entry:
                if (expire := entry[1]) and expire < monotonic():
                    # if still running let it complete if able
                    del self._cached[key]  # continue the same way as if empty
                else:
                    self._cached.move_to_end(key)
                    return entry[0]

        result: Result = self._function(__method_self, *args, **kwargs)  # pyright: ignore[reportUnknownVariableType, reportCallIssue]
        self._cached[key] = _CacheEntry(
            value=result,  # pyright: ignore[reportUnknownArgumentType]
            expire=self._next_expire_time(),
        )
        if len(self._cached) > self._limit:
            # if still running let it complete if able
            self._cached.popitem(last=False)
        return result  # pyright: ignore[reportUnknownArgumentType, reportUnknownVariableType]


class _AsyncCache[**Args, Result]:
    def __init__(
        self,
        function: Callable[Args, Coroutine[None, None, Result]],
        /,
        limit: int,
        expiration: float | None,
    ) -> None:
        self._function: Callable[Args, Coroutine[None, None, Result]] = function
        self._cached: OrderedDict[Hashable, _CacheEntry[Task[Result]]] = OrderedDict()
        self._limit: int = limit
        if expiration := expiration:

            def next_expire_time() -> float | None:
                return monotonic() + expiration
        else:

            def next_expire_time() -> float | None:
                return None

        self._next_expire_time: Callable[[], float | None] = next_expire_time

        # mimic function attributes if able
        mimic_function(function, within=self)

    def __get__(
        self,
        instance: object,
        owner: type | None = None,
        /,
    ) -> Callable[Args, Coroutine[None, None, Result]]:
        if owner is None:
            return self
        else:
            return mimic_function(
                self._function,
                within=partial(
                    self.__method_call__,
                    instance,
                ),
            )

    async def __call__(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        loop: AbstractEventLoop = get_running_loop()
        key: Hashable = _make_key(
            args=args,
            kwds=kwargs,
            typed=True,
        )

        match self._cached.get(key):
            case None:
                pass

            case entry:
                if (expire := entry[1]) and expire < monotonic():
                    # if still running let it complete if able
                    del self._cached[key]  # continue the same way as if empty
                else:
                    self._cached.move_to_end(key)
                    return await shield(entry[0])

        task: Task[Result] = loop.create_task(self._function(*args, **kwargs))  # pyright: ignore[reportCallIssue]
        self._cached[key] = _CacheEntry(
            value=task,
            expire=self._next_expire_time(),
        )
        if len(self._cached) > self._limit:
            # if still running let it complete if able
            self._cached.popitem(last=False)
        return await shield(task)

    async def __method_call__(
        self,
        __method_self: object,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        loop: AbstractEventLoop = get_running_loop()
        key: Hashable = _make_key(
            args=(ref(__method_self), *args),
            kwds=kwargs,
            typed=True,
        )

        match self._cached.get(key):
            case None:
                pass

            case entry:
                if (expire := entry[1]) and expire < monotonic():
                    # if still running let it complete if able
                    del self._cached[key]  # continue the same way as if empty
                else:
                    self._cached.move_to_end(key)
                    return await shield(entry[0])

        task: Task[Result] = loop.create_task(
            self._function(__method_self, *args, **kwargs),  # pyright: ignore[reportCallIssue, reportUnknownArgumentType]
        )
        self._cached[key] = _CacheEntry(
            value=task,
            expire=self._next_expire_time(),
        )
        if len(self._cached) > self._limit:
            # if still running let it complete if able
            self._cached.popitem(last=False)

        return await shield(task)



from asyncio import AbstractEventLoop, CancelledError, Future, get_running_loop
from collections import deque
from collections.abc import AsyncIterator
from typing import Self

from draive.utils import freeze

__all__ = [
    "AsyncQueue",
]


class AsyncQueue[Element](AsyncIterator[Element]):
    def __init__(
        self,
        loop: AbstractEventLoop | None = None,
    ) -> None:
        self._loop: AbstractEventLoop = loop or get_running_loop()
        self._queue: deque[Element] = deque()
        self._waiting: Future[Element] | None = None
        self._finish_reason: BaseException | None = None

        freeze(self)

    def __del__(self) -> None:
        self.finish()

    @property
    def finished(self) -> bool:
        return self._finish_reason is not None

    def enqueue(
        self,
        element: Element,
        /,
        *elements: Element,
    ) -> None:
        if self.finished:
            raise RuntimeError("AsyncQueue is already finished")

        if self._waiting is not None and not self._waiting.done():
            self._waiting.set_result(element)

        else:
            self._queue.append(element)

        self._queue.extend(elements)

    def finish(
        self,
        exception: BaseException | None = None,
    ) -> None:
        if self.finished:
            return  # already finished, ignore

        self._finish_reason = exception or StopAsyncIteration()

        if self._waiting is not None and not self._waiting.done():
            self._waiting.set_exception(self._finish_reason)

    def cancel(self) -> None:
        self.finish(exception=CancelledError())

    def __aiter__(self) -> Self:
        return self

    async def __anext__(self) -> Element:
        assert self._waiting is None, "Only a single queue iterator is supported!"  # nosec: B101

        if self._queue:  # check the queue, let it finish
            return self._queue.popleft()

        if self._finish_reason is not None:  # check if is finished
            raise self._finish_reason

        try:
            # create a new future to wait for next
            self._waiting = self._loop.create_future()
            # wait for the result
            return await self._waiting

        finally:
            # cleanup
            self._waiting = None



from draive.lmm.tools.errors import ToolException
from draive.lmm.tools.specification import ToolSpecification
from draive.lmm.tools.status import ToolContext, ToolStatus
from draive.lmm.tools.tool import AnyTool, Tool, ToolAvailabilityCheck, tool
from draive.lmm.tools.toolbox import Toolbox

__all__ = [
    "AnyTool",
    "tool",
    "Tool",
    "ToolAvailabilityCheck",
    "Toolbox",
    "ToolContext",
    "ToolException",
    "ToolSpecification",
    "ToolStatus",
]



from pathlib import Path
from uuid import UUID

from draive.instructions.errors import MissingInstruction
from draive.instructions.types import Instruction, InstructionFetching
from draive.parameters import DataModel
from draive.utils import asynchronous

__all__ = [
    "instructions_file",
]


def instructions_file(
    path: Path | str,
) -> InstructionFetching:
    repository: _InstructionsFileStorage = _InstructionsFileStorage(path=path)

    async def fetch(
        key: str,
    ) -> Instruction:
        return await repository.instruction(key)

    return fetch


class _Instruction(DataModel):
    identifier: UUID
    content: str


class _Storage(DataModel):
    instructions: dict[str, _Instruction]


class _InstructionsFileStorage:
    def __init__(
        self,
        path: Path | str,
    ) -> None:
        self._path: Path
        match path:
            case str() as path_str:
                self._path = Path(path_str)

            case path:
                self._path = path

        self._storage: _Storage | None = None

    async def instruction(
        self,
        key: str,
    ) -> Instruction:
        if instruction := (await self.storage).instructions.get(key):
            return Instruction(
                instruction.content,
                identifier=instruction.identifier,
            )
        else:
            raise MissingInstruction(
                "%s does not contain instruction for key %s",
                self._path,
                key,
            )

    @property
    async def storage(self) -> _Storage:
        if cache := self._storage:
            return cache

        else:
            loaded: _Storage = await self.load()
            self._storage = loaded
            return loaded

    async def load(
        self,
    ) -> _Storage:
        return await self._file_load()

    @asynchronous(executor=None)
    def _file_load(
        self,
    ) -> _Storage:
        if self._path.exists():
            with open(self._path, mode="rb") as file:
                return _Storage.from_json(file.read())

        else:
            return _Storage(instructions={})



from draive.mistral.client import MistralClient
from draive.mistral.config import MistralChatConfig, MistralEmbeddingConfig
from draive.mistral.embedding import mistral_embed_text
from draive.mistral.errors import MistralException
from draive.mistral.lmm import mistral_lmm_invocation
from draive.mistral.tokenization import mistral_tokenize_text

__all__ = [
    "mistral_embed_text",
    "mistral_lmm_invocation",
    "mistral_tokenize_text",
    "MistralChatConfig",
    "MistralClient",
    "MistralEmbeddingConfig",
    "MistralException",
]



from asyncio import AbstractEventLoop, Event, Handle, gather, get_running_loop
from types import TracebackType
from typing import Any, Self, final

__all__ = [
    "IdleMonitor",
]


@final
class IdleMonitor:
    def __init__(
        self,
        loop: AbstractEventLoop | None = None,
    ) -> None:
        self._loop: AbstractEventLoop = loop or get_running_loop()
        self._running_tasks: int = 0
        self._idle_status: Event = Event()
        # start as running but switches to idle soon if not entered any task after initializing
        self._scheduled_handle: Handle = self._loop.call_soon(self._idle_status.set)
        self._nested_statuses: list[IdleMonitor] = []

    def enter_task(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self._scheduled_handle.cancel()
        self._idle_status.clear()
        self._running_tasks += 1

    def exit_task(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        assert self._running_tasks > 0, "Imbalanced RunningStatus task enter/exit"  # nosec: B101
        self._running_tasks -= 1
        if self._running_tasks > 0:
            return  # still running

        self._scheduled_handle.cancel()
        self._scheduled_handle = self._loop.call_soon(self._idle_status.set)

    def exit_all(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        if self.idle:
            return  # already idle

        self._running_tasks = 0
        self._scheduled_handle.cancel()
        self._scheduled_handle = self._loop.call_soon(self._idle_status.set)

    def nested(self) -> Self:
        nested: Self = self.__class__()
        self._nested_statuses.append(nested)
        return nested

    def __enter__(self) -> None:
        self.enter_task()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.exit_task()

    @property
    def idle(self) -> bool:
        return self._idle_status.is_set() and all(status.idle for status in self._nested_statuses)

    async def wait_idle(self) -> None:
        while not self.idle:
            await gather(
                self._idle_status.wait(),
                *[status.wait_idle() for status in self._nested_statuses],
            )

        assert (  # nosec: B101
            self._running_tasks == 0
        ), "Invalid RunningStatus - notified idle with running tasks"



from collections.abc import Sequence

from draive.conversation.completion import ConversationCompletion
from draive.conversation.lmm import lmm_conversation_completion
from draive.conversation.model import ConversationMessage
from draive.parameters import State
from draive.types import Memory

__all__: list[str] = [
    "Conversation",
]


class Conversation(State):
    completion: ConversationCompletion = lmm_conversation_completion
    memory: Memory[Sequence[ConversationMessage], ConversationMessage] | None = None



from logging import Logger
from typing import Any, cast

from draive.metrics.reporter import MetricsTraceReport, MetricsTraceReporter
from draive.parameters import ParametrizedData
from draive.utils import is_missing

__all__ = [
    "metrics_log_reporter",
]


def metrics_log_reporter(
    list_items_limit: int | None = None,
    item_character_limit: int | None = None,
) -> MetricsTraceReporter:
    async def reporter(
        trace_id: str,
        logger: Logger,
        report: MetricsTraceReport,
    ) -> None:
        report_log: str | None = _report(
            report.with_combined_metrics(),
            list_items_limit=list_items_limit,
            item_character_limit=item_character_limit,
        )
        logger.info(f"[{trace_id}] Metrics report:\n{report_log or 'N/A'}")

    return reporter


def _report(
    report: MetricsTraceReport,
    list_items_limit: int | None,
    item_character_limit: int | None,
) -> str:
    report_log: str = f"@{report.label}({report.duration:.2f}s):"
    for metric_name, metric in report.metrics.items():
        metric_log: str = ""
        for key, value in vars(metric).items():
            if value_log := _value_report(
                value,
                list_items_limit=list_items_limit,
                item_character_limit=item_character_limit,
            ):
                metric_log += f"\n|  + {key}: {value_log}"

            else:
                continue  # skip missing values

        if not metric_log:
            continue  # skip empty logs

        report_log += f"\n• {metric_name}:{metric_log}"

    for nested in report.nested:
        nested_log: str = _report(
            nested,
            list_items_limit=list_items_limit,
            item_character_limit=item_character_limit,
        ).replace("\n", "\n|  ")

        report_log += f"\n{nested_log}"

    return report_log.strip()


def _state_report(
    value: ParametrizedData,
    /,
    list_items_limit: int | None,
    item_character_limit: int | None,
) -> str | None:
    state_log: str = ""
    for key, element in vars(value).items():
        element_log: str | None = _value_report(
            element,
            list_items_limit=list_items_limit,
            item_character_limit=item_character_limit,
        )
        if element_log:
            state_log += f"\n|  + {key}: {element_log}"
        else:
            continue  # skip empty logs

    if state_log:
        return state_log.replace("\n", "\n|  ")
    else:
        return None  # skip empty logs


def _dict_report(
    value: dict[Any, Any],
    /,
    list_items_limit: int | None,
    item_character_limit: int | None,
) -> str | None:
    dict_log: str = ""
    for key, element in value.items():
        element_log: str | None = _value_report(
            element,
            list_items_limit=list_items_limit,
            item_character_limit=item_character_limit,
        )
        if element_log:
            dict_log += f"\n|  + {key}: {element_log}"
        else:
            continue  # skip empty logs

    if dict_log:
        return dict_log.replace("\n", "\n|  ")
    else:
        return None  # skip empty logs


def _list_report(
    value: list[Any],
    /,
    list_items_limit: int | None,
    item_character_limit: int | None,
) -> str | None:
    list_log: str = ""
    enumerared: list[tuple[int, Any]] = list(enumerate(value))
    if list_items_limit:
        if list_items_limit > 0:
            enumerared = enumerared[:list_items_limit]
        else:
            enumerared = enumerared[list_items_limit:]

    for idx, element in enumerared:
        element_log: str | None = _value_report(
            element,
            list_items_limit=list_items_limit,
            item_character_limit=item_character_limit,
        )
        if element_log:
            list_log += f"\n|  [{idx}] {element_log}"
        else:
            continue  # skip empty logs

    if list_log:
        return list_log.replace("\n", "\n|  ")
    else:
        return None  # skip empty logs


def _raw_value_report(
    value: Any,
    /,
    list_items_limit: int | None,
    item_character_limit: int | None,
) -> str | None:
    if is_missing(value):
        return None  # skip missing

    # workaround for pydantic models
    if hasattr(value, "model_dump") and callable(value.model_dump):
        return _dict_report(
            value.model_dump(),
            list_items_limit=list_items_limit,
            item_character_limit=item_character_limit,
        )

    else:
        value_log = str(value)
        if not value_log:
            return None  # skip empty logs

        if (item_character_limit := item_character_limit) and len(value_log) > item_character_limit:
            return value_log.replace("\n", " ")[:item_character_limit] + "..."

        else:
            return value_log.replace("\n", "\n|  ")


def _value_report(
    value: Any,
    /,
    list_items_limit: int | None,
    item_character_limit: int | None,
) -> str | None:
    # try unpack dicts
    if isinstance(value, dict):
        return _dict_report(
            cast(dict[Any, Any], value),
            list_items_limit=list_items_limit,
            item_character_limit=item_character_limit,
        )

    # try unpack lists
    elif isinstance(value, list):
        return _list_report(
            cast(list[Any], value),
            list_items_limit=list_items_limit,
            item_character_limit=item_character_limit,
        )

    # try unpack state
    elif isinstance(value, ParametrizedData):
        return _state_report(
            value,
            list_items_limit=list_items_limit,
            item_character_limit=item_character_limit,
        )

    else:
        return _raw_value_report(
            value,
            list_items_limit=list_items_limit,
            item_character_limit=item_character_limit,
        )



from collections.abc import Sequence
from typing import Any, Literal, overload

from draive.instructions import Instruction
from draive.lmm.invocation import LMMOutputStream, LMMToolSelection
from draive.lmm.state import LMM
from draive.lmm.tools import ToolSpecification
from draive.scope import ctx
from draive.types import LMMContextElement, LMMOutput

__all__ = [
    "lmm_invocation",
]


@overload
async def lmm_invocation(
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: Literal[True],
    **extra: Any,
) -> LMMOutputStream: ...


@overload
async def lmm_invocation(
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: Literal[False] = False,
    **extra: Any,
) -> LMMOutput: ...


@overload
async def lmm_invocation(
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: bool,
    **extra: Any,
) -> LMMOutputStream | LMMOutput: ...


async def lmm_invocation(  # noqa: PLR0913
    *,
    instruction: Instruction | str,
    context: Sequence[LMMContextElement],
    tools: Sequence[ToolSpecification] | None = None,
    tool_selection: LMMToolSelection = "auto",
    output: Literal["text", "json"] = "text",
    stream: bool = False,
    **extra: Any,
) -> LMMOutputStream | LMMOutput:
    return await ctx.state(LMM).invocation(
        instruction=instruction,
        context=context,
        tool_selection=tool_selection,
        tools=tools,
        output=output,
        stream=stream,
        **extra,
    )



from draive.generation.image.call import generate_image
from draive.generation.image.generator import ImageGenerator
from draive.generation.image.state import ImageGeneration

__all__ = [
    "generate_image",
    "ImageGeneration",
    "ImageGenerator",
]



from collections.abc import Sequence
from typing import Literal, overload

from draive.conversation.model import ConversationMessage, ConversationResponseStream
from draive.conversation.state import Conversation
from draive.instructions import Instruction
from draive.lmm import AnyTool, Toolbox
from draive.scope import ctx
from draive.types import Memory, MultimodalContent, MultimodalContentConvertible

__all__ = [
    "conversation_completion",
]


@overload
async def conversation_completion(
    *,
    instruction: Instruction | str,
    input: ConversationMessage | MultimodalContent | MultimodalContentConvertible,
    memory: Memory[Sequence[ConversationMessage], ConversationMessage]
    | Sequence[ConversationMessage]
    | None = None,
    tools: Toolbox | Sequence[AnyTool] | None = None,
    stream: Literal[True],
) -> ConversationResponseStream: ...


@overload
async def conversation_completion(
    *,
    instruction: Instruction | str,
    input: ConversationMessage | MultimodalContent | MultimodalContentConvertible,
    memory: Memory[Sequence[ConversationMessage], ConversationMessage]
    | Sequence[ConversationMessage]
    | None = None,
    tools: Toolbox | Sequence[AnyTool] | None = None,
    stream: Literal[False] = False,
) -> ConversationMessage: ...


@overload
async def conversation_completion(
    *,
    instruction: Instruction | str,
    input: ConversationMessage | MultimodalContent | MultimodalContentConvertible,
    memory: Memory[Sequence[ConversationMessage], ConversationMessage]
    | Sequence[ConversationMessage]
    | None = None,
    tools: Toolbox | Sequence[AnyTool] | None = None,
    stream: bool,
) -> ConversationResponseStream | ConversationMessage: ...


async def conversation_completion(
    *,
    instruction: Instruction | str,
    input: ConversationMessage | MultimodalContent | MultimodalContentConvertible,  # noqa: A002
    memory: Memory[Sequence[ConversationMessage], ConversationMessage]
    | Sequence[ConversationMessage]
    | None = None,
    tools: Toolbox | Sequence[AnyTool] | None = None,
    stream: bool = False,
) -> ConversationResponseStream | ConversationMessage:
    conversation: Conversation = ctx.state(Conversation)
    return await conversation.completion(
        instruction=instruction,
        input=input,
        memory=memory or conversation.memory,
        tools=tools,
        stream=stream,
    )



from asyncio import CancelledError
from collections.abc import AsyncIterable, Generator
from concurrent.futures import Executor, ThreadPoolExecutor
from typing import Final

import pytest
from draive import ctx
from pytest import raises

EXECUTOR: Final[Executor | None] = ThreadPoolExecutor(max_workers=1)


class FakeException(Exception):
    pass


@pytest.mark.asyncio
@ctx.wrap("test")
async def test_fails_when_streaming_fails():
    def stream_job() -> Generator[int]:
        raise FakeException()

    elements: int = 0
    with raises(FakeException):
        async for _ in ctx.stream_sync(stream_job(), executor=EXECUTOR):
            elements += 1

    assert elements == 0


@pytest.mark.asyncio
@ctx.wrap("test")
async def test_cancels_when_streaming_cancels():
    def stream_job() -> Generator[int]:
        while False:
            yield  # finish without updates
        raise CancelledError()

    elements: int = 0
    with raises(CancelledError):
        stream: AsyncIterable[int] = ctx.stream_sync(stream_job(), executor=EXECUTOR)
        async for _ in stream:
            elements += 1

    assert elements == 0


@pytest.mark.asyncio
@ctx.wrap("test")
async def test_cancels_when_task_cancels():
    def stream_job() -> Generator[int]:
        while False:
            yield  # finish without updates

    elements: int = 0
    with raises(CancelledError):
        stream: AsyncIterable[int] = ctx.stream_sync(stream_job(), executor=EXECUTOR)
        ctx.cancel()
        async for _ in stream:
            elements += 1

    assert elements == 0


@pytest.mark.asyncio
@ctx.wrap("test")
async def test_ends_when_iteration_ends():
    def stream_job() -> Generator[int]:
        while False:
            yield  # finish without updates

    elements: int = 0
    async for _ in ctx.stream_sync(stream_job(), executor=EXECUTOR):
        elements += 1

    assert elements == 0


@pytest.mark.asyncio
@ctx.wrap("test")
async def test_delivers_values_from_generator():
    def stream_job() -> Generator[int]:
        yield 0
        yield 1
        yield 2
        yield 3

    stream: AsyncIterable[int] = ctx.stream_sync(stream_job(), executor=EXECUTOR)
    elements: list[int] = []

    async for element in stream:
        elements.append(element)

    assert elements == [0, 1, 2, 3]


@pytest.mark.asyncio
@ctx.wrap("test")
async def test_delivers_errors_from_generator():
    def stream_job() -> Generator[int]:
        yield 0
        yield 1
        yield 2

        raise FakeException()

        yield 3

    stream: AsyncIterable[int] = ctx.stream_sync(stream_job(), executor=EXECUTOR)
    elements: list[int] = []

    with raises(FakeException):
        async for element in stream:
            elements.append(element)

    assert elements == [0, 1, 2]



from draive.evaluation import EvaluationScore, evaluator
from draive.evaluators.score import CommonScoreModel
from draive.generation import generate_model

__all__ = [
    "text_fluency_evaluator",
]


INSTRUCTION: str = """\
You will be given a text. Your task is to rate this text using only the Fluency metric, \
which is described in the Evaluation Criteria.
Please make sure you read and understand these instructions very carefully.
Keep this document open while reviewing, and refer to it as needed.

Evaluation Criteria:
Fluency (0.0-2.0) - the quality of the text in terms of grammar, spelling, punctuation, \
word choice, and sentence structure.

Rating Scale:
0.0: Poor - the text has many errors that make it hard to understand or sound unnatural.
1.0: Fair - the text has some errors that affect the clarity or smoothness of the text, \
but the main points are still comprehensible.
2.0: Good - the text has few or no errors and is easy to read and follow.

Evaluation Steps:
1. Read the text and evaluate its fluency based on the given criteria.
2. Assign a fluency score from 0.0 to 2.0 based on the provided criteria.

Important: The score must be a decimal number from 0.0 to 2.0. 2.0 is the maximum, \
do not exceed this value.
"""


INPUT_TEMPLATE: str = """
<TEXT>
{text}
</TEXT>
"""


@evaluator(name="text_fluency")
async def text_fluency_evaluator(
    text: str,
    /,
) -> EvaluationScore:
    if not text:
        return EvaluationScore(
            value=0,
            comment="Input text was empty!",
        )

    score: CommonScoreModel = await generate_model(
        CommonScoreModel,
        instruction=INSTRUCTION,
        input=INPUT_TEMPLATE.format(text=text),
        examples=[
            (
                INPUT_TEMPLATE.format(
                    text=(
                        "The cat sitted on mat. It were very comfrotable. "
                        "The sun shine bright in sky."
                    ),
                ),
                CommonScoreModel(score=0.0),
            ),
            (
                INPUT_TEMPLATE.format(
                    text=(
                        "The movie was good, but I didn't liked the ending. "
                        "It left me feeling confuse and unsatisfied."
                    ),
                ),
                CommonScoreModel(score=1.0),
            ),
            (
                INPUT_TEMPLATE.format(
                    text=(
                        "The concert last night was amazing. "
                        "The band played all their hit songs, and the crowd was energetic "
                        "throughout the performance."
                    ),
                ),
                CommonScoreModel(score=2.0),
            ),
        ],
    )

    return score.normalized(divider=2)



from draive import DataModel

__all__ = [
    "FastembedTextConfig",
    "FastembedImageConfig",
]


class FastembedTextConfig(DataModel):
    model: str = "nomic-ai/nomic-embed-text-v1.5"
    cache_dir: str | None = "./embedding_models/"


class FastembedImageConfig(DataModel):
    model: str = "Qdrant/resnet50-onnx"
    cache_dir: str | None = "./embedding_models/"



from collections.abc import Iterable, Sequence
from typing import Any, Literal

from draive.generation.model.generator import ModelGeneratorDecoder
from draive.instructions import Instruction
from draive.lmm import AnyTool, Toolbox, lmm_invocation
from draive.parameters import DataModel
from draive.scope import ctx
from draive.types import (
    LMMCompletion,
    LMMContextElement,
    LMMInput,
    LMMToolRequests,
    LMMToolResponse,
    MultimodalContent,
    MultimodalContentConvertible,
)

__all__: list[str] = [
    "lmm_generate_model",
]


async def lmm_generate_model[Generated: DataModel](  # noqa: PLR0913, C901, PLR0912
    generated: type[Generated],
    /,
    *,
    instruction: Instruction | str,
    input: MultimodalContent | MultimodalContentConvertible,  # noqa: A002
    schema_injection: Literal["auto", "full", "simplified", "skip"] = "auto",
    tools: Toolbox | Sequence[AnyTool] | None = None,
    examples: Iterable[tuple[MultimodalContent | MultimodalContentConvertible, Generated]]
    | None = None,
    decoder: ModelGeneratorDecoder | None = None,
    **extra: Any,
) -> Generated:
    with ctx.nested("lmm_generate_model"):
        toolbox: Toolbox
        match tools:
            case None:
                toolbox = Toolbox()

            case Toolbox() as tools:
                toolbox = tools

            case [*tools]:
                toolbox = Toolbox(*tools)

        generation_instruction: Instruction
        match instruction:
            case str(instruction):
                generation_instruction = Instruction(instruction)

            case Instruction() as instruction:
                generation_instruction = instruction

        extended_instruction: Instruction
        match schema_injection:
            case "auto":
                extended_instruction = Instruction.of(
                    generation_instruction.extended(
                        DEFAULT_INSTRUCTION_EXTENSION.format(
                            schema=generated.simplified_schema(indent=2),
                        ),
                        joiner="\n\n",
                    ),
                )

            case "full":
                extended_instruction = Instruction.of(
                    generation_instruction,
                    schema=generated.json_schema(indent=2),
                )

            case "simplified":
                extended_instruction = Instruction.of(
                    generation_instruction,
                    schema=generated.simplified_schema(indent=2),
                )

            case "skip":
                extended_instruction = Instruction.of(
                    generation_instruction,
                )

        context: list[LMMContextElement] = [
            *[
                message
                for example in examples or []
                for message in [
                    LMMInput.of(example[0]),
                    LMMCompletion.of(example[1].as_json(indent=2)),
                ]
            ],
            LMMInput.of(input),
        ]

        recursion_level: int = 0
        while recursion_level <= toolbox.recursion_limit:
            match await lmm_invocation(
                instruction=extended_instruction,
                context=context,
                tools=toolbox.available_tools(),
                tool_selection=toolbox.tool_selection(recursion_level=recursion_level),
                output="json",
                stream=False,
                **extra,
            ):
                case LMMCompletion() as completion:
                    ctx.log_debug("Received model generation result")
                    if decoder := decoder:
                        return generated.from_dict(decoder(completion.content))

                    else:
                        return generated.from_json(completion.content.as_string())

                case LMMToolRequests() as tool_requests:
                    ctx.log_debug("Received model generation tool calls")
                    context.append(tool_requests)
                    responses: list[LMMToolResponse] = await toolbox.respond(tool_requests)

                    if direct_responses := [response for response in responses if response.direct]:
                        for response in direct_responses:
                            if isinstance(response, generated):
                                # return first response matching requested model
                                return response

                            else:
                                continue

                        direct_responses_content: MultimodalContent = MultimodalContent.of(
                            *[response.content for response in direct_responses]
                        )

                        # TODO: check if this join makes any sense,
                        # perhaps we could merge json objects instead?
                        if decoder := decoder:
                            return generated.from_dict(decoder(direct_responses_content))

                        else:
                            return generated.from_json(direct_responses_content.as_string())

                    else:
                        context.extend(responses)

            recursion_level += 1  # continue with next recursion level

        raise RuntimeError("LMM exceeded limit of recursive calls")


DEFAULT_INSTRUCTION_EXTENSION: str = """\
The output have to be a JSON conforming to the following schema:
```
{schema}
```
Provide ONLY a single, raw, valid JSON without any comments, formatting or additional elements.
"""



from draive.mrs.client import MRSClient
from draive.mrs.config import MRSChatConfig
from draive.mrs.errors import MRSException
from draive.mrs.lmm import mrs_lmm_invocation

__all__ = [
    "mrs_lmm_invocation",
    "MRSChatConfig",
    "MRSClient",
    "MRSException",
]



from asyncio import iscoroutinefunction
from collections.abc import Callable, Coroutine
from typing import Any, cast

from draive.metrics import ArgumentsTrace, ResultTrace
from draive.scope import ctx
from draive.utils import mimic_function

__all__ = [
    "traced",
]


def traced[**Args, Result](
    function: Callable[Args, Result],
    /,
) -> Callable[Args, Result]:
    if iscoroutinefunction(function):
        return cast(
            Callable[Args, Result],
            _traced_async(function),
        )
    else:
        return _traced_sync(function)


def _traced_sync[**Args, Result](
    function: Callable[Args, Result],
    /,
) -> Callable[Args, Result]:
    label: str = function.__name__

    @mimic_function(function)
    def wrapped(
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        with ctx.nested(
            label,
            metrics=[ArgumentsTrace.of(*args, **kwargs)],
        ):
            result: Result = function(*args, **kwargs)
            ctx.record(ResultTrace.of(result))
            return result

    return wrapped


def _traced_async[**Args, Result](
    function: Callable[Args, Coroutine[Any, Any, Result]],
    /,
) -> Callable[Args, Coroutine[Any, Any, Result]]:
    label: str = function.__name__

    @mimic_function(function)
    async def wrapped(
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        with ctx.nested(
            label,
            metrics=[ArgumentsTrace.of(*args, **kwargs)],
        ):
            result: Result = await function(*args, **kwargs)
            ctx.record(ResultTrace.of(result))
            return result

    return wrapped



from typing import Protocol, Self, final, runtime_checkable
from uuid import UUID, uuid4

from draive.utils import freeze

__all__ = [
    "Instruction",
    "InstructionFetching",
]


@final
class Instruction:
    @classmethod
    def of(
        cls,
        instruction: Self | str,
        /,
        **variables: object,
    ) -> Self:
        match instruction:
            case str() as content:
                return cls(
                    content,
                    identifier=None,
                    **variables,
                )

            case instruction:
                return instruction.updated(**variables)

    def __init__(
        self,
        instruction: str,
        /,
        identifier: UUID | None = None,
        **variables: object,
    ) -> None:
        self.instruction: str = instruction
        self.identifier: UUID = identifier or uuid4()
        self.variables: dict[str, object] = variables

        freeze(self)

    def format(
        self,
        **variables: object,
    ) -> str:
        if variables:
            return self.instruction.format_map(
                {
                    **self.variables,
                    **variables,
                },
            )

        elif self.variables:
            return self.instruction.format_map(self.variables)

        else:
            return self.instruction

    def extended(
        self,
        instruction: str,
        /,
        joiner: str | None = None,
        **variables: object,
    ) -> Self:
        if variables:
            return self.__class__(
                (joiner or " ").join((self.instruction, instruction)),
                identifier=self.identifier,
                **{
                    **self.variables,
                    **variables,
                },
            )

        else:
            return self.__class__(
                (joiner or " ").join((self.instruction, instruction)),
                identifier=self.identifier,
                **self.variables,
            )

    def updated(
        self,
        **variables: object,
    ) -> Self:
        if variables:
            return self.__class__(
                self.instruction,
                identifier=self.identifier,
                **{
                    **self.variables,
                    **variables,
                },
            )

        else:
            return self

    def __str__(self) -> str:
        try:
            return self.format()

        except KeyError:
            return self.instruction


@runtime_checkable
class InstructionFetching(Protocol):
    async def __call__(
        self,
        key: str,
    ) -> Instruction: ...



from typing import Literal, TypedDict

from draive.parameters import DataModel
from draive.utils import MISSING, Missing

__all__ = [
    "AnthropicConfig",
]


class ResponseFormat(TypedDict):
    type: Literal["text", "json_object"]


class AnthropicConfig(DataModel):
    model: (
        Literal[
            "claude-3-5-sonnet-20240620",
            "claude-3-opus-20240229",
            "claude-3-sonnet-20240229",
            "claude-3-haiku-20240307",
        ]
        | str
    ) = "claude-3-haiku-20240307"
    temperature: float = 0.0
    top_p: float | Missing = MISSING
    max_tokens: int = 2048
    timeout: float | Missing = MISSING



from collections.abc import Callable, Coroutine
from typing import Any

__all__ = [
    "always",
    "async_always",
]


def always[Value](
    value: Value,
    /,
) -> Callable[..., Value]:
    def always_value(
        *args: Any,
        **kwargs: Any,
    ) -> Value:
        return value

    return always_value


def async_always[Value](
    value: Value,
    /,
) -> Callable[..., Coroutine[None, None, Value]]:
    async def always_value(
        *args: Any,
        **kwargs: Any,
    ) -> Value:
        return value

    return always_value



from draive.conversation.call import conversation_completion
from draive.conversation.completion import ConversationCompletion
from draive.conversation.lmm import lmm_conversation_completion
from draive.conversation.model import (
    ConversationMessage,
    ConversationMessageChunk,
    ConversationResponseStream,
)
from draive.conversation.state import Conversation

__all__ = [
    "conversation_completion",
    "Conversation",
    "ConversationCompletion",
    "ConversationMessageChunk",
    "ConversationResponseStream",
    "ConversationMessage",
    "lmm_conversation_completion",
]



import builtins
import datetime
import enum
import types
import typing
import uuid
from collections import abc as collections_abc
from collections.abc import Callable, Sequence
from dataclasses import is_dataclass
from typing import Any

import draive.utils as draive_utils
from draive.parameters.annotations import resolve_annotation
from draive.parameters.errors import ParameterValidationContext, ParameterValidationError

__all__ = [
    "as_validator",
    "parameter_validator",
    "ParameterValidator",
    "ParameterVerifier",
]


type ParameterValidator[Value] = Callable[[Any, ParameterValidationContext], Value]
type ParameterVerifier[Value] = Callable[[Value], None]


def as_validator[Value](
    function: Callable[[Any], Value],
) -> ParameterValidator[Value]:
    def validator(
        value: Any,
        context: ParameterValidationContext,
    ) -> Value:
        try:
            return function(value)

        except Exception as exc:
            raise ParameterValidationError.invalid(
                exception=exc,
                context=context,
            ) from exc

    return validator


def _none_validator(
    value: Any,
    context: ParameterValidationContext,
) -> Any:
    match value:
        case None:
            return None

        case _:
            raise ParameterValidationError.invalid_type(
                expected=types.NoneType,
                received=value,
                context=context,
            )


def _missing_validator(
    value: Any,
    context: ParameterValidationContext,
) -> Any:
    match value:
        case missing if isinstance(missing, draive_utils.Missing):
            return missing

        case None:
            return draive_utils.MISSING

        case _:
            raise ParameterValidationError.invalid_type(
                expected=draive_utils.Missing,
                received=value,
                context=context,
            )


def _any_validator(
    value: Any,
    context: ParameterValidationContext,
) -> Any:
    return value  # any is always valid


def _str_validator(
    value: Any,
    context: ParameterValidationContext,
) -> Any:
    match value:
        case str() as value:
            return value

        case _:
            raise ParameterValidationError.invalid_type(
                expected=str,
                received=value,
                context=context,
            )


def _int_validator(
    value: Any,
    context: ParameterValidationContext,
) -> Any:
    match value:
        case int() as value:
            return value

        case _:
            raise ParameterValidationError.invalid_type(
                expected=int,
                received=value,
                context=context,
            )


def _float_validator(
    value: Any,
    context: ParameterValidationContext,
) -> Any:
    match value:
        case float() as value:
            return value

        case int() as convertible:
            return float(convertible)

        case _:
            raise ParameterValidationError.invalid_type(
                expected=float,
                received=value,
                context=context,
            )


def _bool_validator(
    value: Any,
    context: ParameterValidationContext,
) -> Any:
    match value:
        case bool() as value:
            return value

        case int() as convertible:
            return bool(convertible != 0)

        case str() as convertible if convertible.lower() == "true":
            return True

        case str() as convertible if convertible.lower() == "false":
            return False

        case _:
            raise ParameterValidationError.invalid_type(
                expected=bool,
                received=value,
                context=context,
            )


def _prepare_tuple_validator(
    elements_annotation: tuple[Any, ...],
    /,
    type_arguments: dict[str, Any],
    globalns: dict[str, Any] | None,
    localns: dict[str, Any] | None,
    recursion_guard: frozenset[type[Any]],
) -> ParameterValidator[Any]:
    match elements_annotation:
        case [element, builtins.Ellipsis]:
            element_validator: ParameterValidator[Any] = parameter_validator(
                element,
                verifier=None,
                type_arguments=type_arguments,
                globalns=globalns,
                localns=localns,
                recursion_guard=recursion_guard,
            )

            def variable_tuple_validator(
                value: Any,
                context: ParameterValidationContext,
            ) -> Any:
                match value:
                    case [*values] as value:
                        return tuple[Any, ...](
                            element_validator(element, context.appending_path(f"[{idx}]"))
                            for idx, element in enumerate(values)
                        )

                    case []:
                        return tuple[Any, ...]()

                    case _:
                        raise ParameterValidationError.invalid_type(
                            expected=tuple,
                            received=value,
                            context=context,
                        )

            return variable_tuple_validator

        case [*fixed]:
            element_validators: list[ParameterValidator[Any]] = [
                parameter_validator(
                    element,
                    verifier=None,
                    type_arguments=type_arguments,
                    globalns=globalns,
                    localns=localns,
                    recursion_guard=recursion_guard,
                )
                for element in fixed
            ]

            def fixed_tuple_validator(
                value: Any,
                context: ParameterValidationContext,
            ) -> Any:
                match value:
                    case [*values] as value if len(values) == len(element_validators):
                        return tuple[Any, ...](
                            validator(element, context.appending_path(f"[{idx}]"))
                            for (idx, element), validator in zip(
                                enumerate(values),
                                element_validators,
                                strict=True,
                            )
                        )

                    case _:
                        raise ParameterValidationError.invalid_type(
                            expected=tuple,
                            received=value,
                            context=context,
                        )

            return fixed_tuple_validator


def _prepare_list_validator(
    elements_annotation: tuple[Any, ...],
    /,
    type_arguments: dict[str, Any],
    globalns: dict[str, Any] | None,
    localns: dict[str, Any] | None,
    recursion_guard: frozenset[type[Any]],
) -> ParameterValidator[Any]:
    element_validator: ParameterValidator[Any]
    match elements_annotation:
        case [element]:
            element_validator = parameter_validator(
                element,
                verifier=None,
                type_arguments=type_arguments,
                globalns=globalns,
                localns=localns,
                recursion_guard=recursion_guard,
            )

        case other:
            raise TypeError(f"Unsupported annotation - list[{other}]")

    def list_validator(
        value: Any,
        context: ParameterValidationContext,
    ) -> Any:
        match value:
            case [*values] as value:
                return list[Any](
                    element_validator(element, context.appending_path(f"[{idx}]"))
                    for idx, element in enumerate(values)
                )

            case []:
                return list[Any]()

            case _:
                raise ParameterValidationError.invalid_type(
                    expected=list,
                    received=value,
                    context=context,
                )

    return list_validator


def _prepare_set_validator(
    elements_annotation: tuple[Any, ...],
    /,
    type_arguments: dict[str, Any],
    globalns: dict[str, Any] | None,
    localns: dict[str, Any] | None,
    recursion_guard: frozenset[type[Any]],
) -> ParameterValidator[Any]:
    element_validator: ParameterValidator[Any]
    match elements_annotation:
        case [element]:
            element_validator = parameter_validator(
                element,
                verifier=None,
                type_arguments=type_arguments,
                globalns=globalns,
                localns=localns,
                recursion_guard=recursion_guard,
            )

        case other:
            raise TypeError(f"Unsupported annotation - set[{other}]")

    def set_validator(
        value: Any,
        context: ParameterValidationContext,
    ) -> Any:
        match value:
            case [*values] as value:
                return set[Any](
                    element_validator(element, context.appending_path(f"[{idx}]"))
                    for idx, element in enumerate(values)
                )

            case []:
                return set[Any]()

            case _:
                raise ParameterValidationError.invalid_type(
                    expected=list,
                    received=value,
                    context=context,
                )

    return set_validator


def _prepare_union_validator(
    elements_annotation: tuple[Any, ...],
    /,
    type_arguments: dict[str, Any],
    globalns: dict[str, Any] | None,
    localns: dict[str, Any] | None,
    recursion_guard: frozenset[type[Any]],
) -> ParameterValidator[Any]:
    alternatives_validators: list[ParameterValidator[Any]] = [
        parameter_validator(
            alternative,
            verifier=None,
            type_arguments=type_arguments,
            globalns=globalns,
            localns=localns,
            recursion_guard=recursion_guard,
        )
        for alternative in elements_annotation
    ]

    def union_validator(
        value: Any,
        context: ParameterValidationContext,
    ) -> Any:
        for validator in alternatives_validators:
            try:
                return validator(value, context)

            except ParameterValidationError:
                continue

        raise ParameterValidationError.invalid_union_type(
            expected=elements_annotation,
            received=value,
            context=context,
        )

    return union_validator


def _prepare_str_enum_validator(
    values: Sequence[str],
    /,
) -> ParameterValidator[Any]:
    def str_enum_validator(
        value: Any,
        context: ParameterValidationContext,
    ) -> Any:
        match value:
            case str() as value:
                if value in values:
                    return value

                else:
                    raise ParameterValidationError.invalid_value(
                        expected=values,
                        received=value,
                        context=context,
                    )

            case _:
                raise ParameterValidationError.invalid_type(
                    expected=str,
                    received=value,
                    context=context,
                )

    return str_enum_validator


def _prepare_int_enum_validator(
    values: Sequence[int],
    /,
) -> ParameterValidator[Any]:
    def int_enum_validator(
        value: Any,
        context: ParameterValidationContext,
    ) -> Any:
        match value:
            case int() as value:
                if value in values:
                    return value

                else:
                    raise ParameterValidationError.invalid_value(
                        expected=values,
                        received=value,
                        context=context,
                    )

            case _:
                raise ParameterValidationError.invalid_type(
                    expected=int,
                    received=value,
                    context=context,
                )

    return int_enum_validator


def _prepare_any_enum_validator(
    values: Sequence[Any],
    /,
) -> ParameterValidator[Any]:
    def enum_validator(
        value: Any,
        context: ParameterValidationContext,
    ) -> Any:
        if value in values:
            return value

        else:
            raise ParameterValidationError.invalid_value(
                expected=values,
                received=value,
                context=context,
            )

    return enum_validator


def _uuid_validator(
    value: Any,
    context: ParameterValidationContext,
) -> Any:
    match value:
        case str() as value:
            try:
                return uuid.UUID(hex=value)

            except Exception as exc:
                raise ParameterValidationError.invalid_value(
                    expected=uuid.UUID,
                    received=value,
                    context=context,
                ) from exc

        case uuid.UUID() as value:
            return value

        case bytes() as value:
            try:
                return uuid.UUID(bytes=value)

            except Exception as exc:
                raise ParameterValidationError.invalid_value(
                    expected=uuid.UUID,
                    received=value,
                    context=context,
                ) from exc

        case _:
            raise ParameterValidationError.invalid_type(
                expected=uuid.UUID,
                received=value,
                context=context,
            )


def _datetime_validator(
    value: Any,
    context: ParameterValidationContext,
) -> Any:
    match value:
        case str() as value:
            try:
                return datetime.datetime.fromisoformat(value)

            except Exception as exc:
                raise ParameterValidationError.invalid_value(
                    expected=datetime.datetime,
                    received=value,
                    context=context,
                ) from exc

        case datetime.datetime() as value:
            return value

        case int() as value:
            try:
                return datetime.datetime.fromtimestamp(
                    value,
                    datetime.UTC,
                )

            except Exception as exc:
                raise ParameterValidationError.invalid_value(
                    expected=datetime.datetime,
                    received=value,
                    context=context,
                ) from exc

        case float() as value:
            try:
                return datetime.datetime.fromtimestamp(
                    value,
                    datetime.UTC,
                )

            except Exception as exc:
                raise ParameterValidationError.invalid_value(
                    expected=datetime.datetime,
                    received=value,
                    context=context,
                ) from exc

        case _:
            raise ParameterValidationError.invalid_type(
                expected=datetime.datetime,
                received=value,
                context=context,
            )


def _callable_validator(
    value: Any,
    context: ParameterValidationContext,
) -> Any:
    match value:
        # TODO: verify signature eventually
        case value if callable(value):
            return value

        case _:
            raise ParameterValidationError.invalid_type(
                expected=collections_abc.Callable,
                received=value,
                context=context,
            )


def _prepare_type_validator(
    validated_type: type[Any],
    /,
) -> ParameterValidator[Any]:
    def type_validator(
        value: Any,
        context: ParameterValidationContext,
    ) -> Any:
        match value:
            case value if isinstance(value, validated_type):
                return value

            case _:
                raise ParameterValidationError.invalid_type(
                    expected=validated_type,
                    received=value,
                    context=context,
                )

    return type_validator


def _prepare_meta_type_validator(
    validated_type: type[Any],
    /,
) -> ParameterValidator[Any]:
    def type_validator(
        value: Any,
        context: ParameterValidationContext,
    ) -> Any:
        match value:
            case type() as type_value:
                if type_value == validated_type:
                    return type_value

                else:
                    raise ParameterValidationError.invalid_value(
                        expected=validated_type,
                        received=type_value,
                        context=context,
                    )

            case _:
                raise ParameterValidationError.invalid_type(
                    expected=type,
                    received=value,
                    context=context,
                )

    return type_validator


def _prepare_dataclass_validator(
    dataclass_type: type[Any],
    /,
) -> ParameterValidator[Any]:
    def dataclass_validator(
        value: Any,
        context: ParameterValidationContext,
    ) -> Any:
        match value:
            case value if isinstance(value, dataclass_type):
                return value

            case {**parameters}:
                # TODO: nested validation?
                try:
                    return dataclass_type(**parameters)

                except Exception as exc:
                    raise ParameterValidationError.invalid_value(
                        expected=dataclass_type,
                        received=value,
                        context=context,
                    ) from exc

            case _:
                raise ParameterValidationError.invalid_type(
                    expected=dataclass_type,
                    received=value,
                    context=context,
                )

    return dataclass_validator


def _prepare_dict_validator(
    items_annotation: tuple[Any, ...],
    /,
    type_arguments: dict[str, Any],
    globalns: dict[str, Any] | None,
    localns: dict[str, Any] | None,
    recursion_guard: frozenset[type[Any]],
) -> ParameterValidator[Any]:
    key_validator: ParameterValidator[Any]
    value_validator: ParameterValidator[Any]
    match items_annotation:
        case [key, value]:
            key_validator = parameter_validator(
                key,
                verifier=None,
                type_arguments=type_arguments,
                globalns=globalns,
                localns=localns,
                recursion_guard=recursion_guard,
            )
            value_validator = parameter_validator(
                value,
                verifier=None,
                type_arguments=type_arguments,
                globalns=globalns,
                localns=localns,
                recursion_guard=recursion_guard,
            )

        case other:
            raise TypeError(f"Unsupported annotation - dict[{other}]")

    def dict_validator(
        value: Any,
        context: ParameterValidationContext,
    ) -> Any:
        match value:
            case {**values}:
                return {
                    key_validator(key, context): value_validator(
                        value, context.appending_path(f"[{key}]")
                    )
                    for key, value in values.items()
                }

            case _:
                raise ParameterValidationError.invalid_type(
                    expected=dict,
                    received=value,
                    context=context,
                )

    return dict_validator


def _prepare_typed_dict_validator(
    typed_dict_annotation: type[Any],
    /,
    type_arguments: dict[str, Any],
    globalns: dict[str, Any] | None,
    localns: dict[str, Any] | None,
    recursion_guard: frozenset[type[Any]],
) -> ParameterValidator[Any]:
    required_fields: set[str] = typed_dict_annotation.__required_keys__
    field_validators: dict[str, ParameterValidator[Any]] = {
        name: parameter_validator(
            annotation,
            verifier=None,
            type_arguments=type_arguments,
            globalns=globalns,
            localns=localns,
            recursion_guard=recursion_guard,
        )
        for name, annotation in typed_dict_annotation.__annotations__.items()
    }

    def typed_dict_validator(
        value: Any,
        context: ParameterValidationContext,
    ) -> Any:
        match value:
            case {**values}:
                result: dict[str, Any] = {}
                for key, item in values.items():
                    match key:
                        case str() as name if name in field_validators:
                            result[key] = field_validators[name](
                                item, context.appending_path(f"[{key}]")
                            )

                        case other:
                            raise ParameterValidationError.invalid_type(
                                expected=str,
                                received=other,
                                context=context,
                            )

                missing_values: set[str] = required_fields.difference(result.keys())
                if missing_values:
                    raise ParameterValidationError.missing(context=context)

                return result

            case _:
                raise ParameterValidationError.invalid_type(
                    expected=dict,
                    received=value,
                    context=context,
                )

    return typed_dict_validator


def _verified[Value](
    validator: ParameterValidator[Value],
    /,
    *,
    verifier: ParameterVerifier[Value],
) -> ParameterValidator[Value]:
    def wrapped(
        value: Any,
        context: ParameterValidationContext,
    ) -> Value:
        validated: Value = validator(value, context)

        try:
            verifier(validated)

        except ParameterValidationError as exc:
            raise exc

        except Exception as exc:
            raise ParameterValidationError.invalid(
                exception=exc,
                context=context,
            ) from exc

        return validated

    return wrapped


def parameter_validator[Value](  # noqa: C901, PLR0915, PLR0912, PLR0913
    annotation: Any,
    /,
    verifier: Callable[[Value], None] | None,
    type_arguments: dict[str, Any],
    globalns: dict[str, Any] | None,
    localns: dict[str, Any] | None,
    recursion_guard: frozenset[type[Any]],
) -> ParameterValidator[Value]:
    resolved_origin, resolved_args = resolve_annotation(
        annotation,
        type_arguments=type_arguments,
        globalns=globalns,
        localns=localns,
    )

    if resolved_origin in recursion_guard:
        return resolved_origin.validator

    validator: ParameterValidator[Value]
    match resolved_origin:
        case builtins.str:
            validator = _str_validator

        case builtins.int:
            validator = _int_validator

        case builtins.float:
            validator = _float_validator

        case builtins.bool:
            validator = _bool_validator

        case types.NoneType:
            validator = _none_validator

        case types.UnionType:
            validator = _prepare_union_validator(
                resolved_args,
                globalns=globalns,
                type_arguments=type_arguments,
                localns=localns,
                recursion_guard=recursion_guard,
            )

        case builtins.tuple:  # pyright: ignore[reportUnknownMemberType]
            validator = _prepare_tuple_validator(
                resolved_args,
                globalns=globalns,
                type_arguments=type_arguments,
                localns=localns,
                recursion_guard=recursion_guard,
            )

        case parametrized if hasattr(parametrized, "__PARAMETERS__"):
            validator = parametrized.validator

        case data_class if is_dataclass(data_class):
            validator = _prepare_dataclass_validator(data_class)

        case typed_dict if typing.is_typeddict(typed_dict):
            validator = _prepare_typed_dict_validator(
                typed_dict,
                globalns=globalns,
                type_arguments=type_arguments,
                localns=localns,
                recursion_guard=recursion_guard,
            )

        case builtins.dict | collections_abc.Mapping:  # pyright: ignore[reportUnknownMemberType]
            validator = _prepare_dict_validator(
                resolved_args,
                type_arguments=type_arguments,
                globalns=globalns,
                localns=localns,
                recursion_guard=recursion_guard,
            )

        case builtins.set | collections_abc.Set:
            validator = _prepare_set_validator(
                resolved_args,
                type_arguments=type_arguments,
                globalns=globalns,
                localns=localns,
                recursion_guard=recursion_guard,
            )

        case builtins.list | collections_abc.Sequence:  # pyright: ignore[reportUnknownMemberType]
            validator = _prepare_list_validator(
                resolved_args,
                type_arguments=type_arguments,
                globalns=globalns,
                localns=localns,
                recursion_guard=recursion_guard,
            )

        case typing.Literal:
            if all(isinstance(option, str) for option in resolved_args):
                validator = _prepare_str_enum_validator(resolved_args)

            elif all(isinstance(option, int) for option in resolved_args):
                validator = _prepare_int_enum_validator(resolved_args)

            else:
                validator = _prepare_any_enum_validator(resolved_args)

        case enum_type if isinstance(enum_type, enum.EnumType):
            if isinstance(enum_type, enum.StrEnum):
                validator = _prepare_str_enum_validator(resolved_args)

            elif isinstance(enum_type, enum.IntEnum):
                validator = _prepare_int_enum_validator(resolved_args)

            else:
                raise TypeError(f"Unsupported enum type annotation: {annotation}")

        case datetime.datetime:
            validator = _datetime_validator

        case uuid.UUID:
            validator = _uuid_validator

        case collections_abc.Callable:  # pyright: ignore[reportUnknownMemberType, reportUnnecessaryComparison]
            validator = _callable_validator

        case draive_utils.Missing:
            validator = _missing_validator

        case builtins.type:
            validator = _prepare_meta_type_validator(annotation)

        case typing.Any:
            validator = _any_validator

        case type() as other_type:
            validator = _prepare_type_validator(other_type)

        case other:
            raise TypeError("Unsupported type annotation: %s", other)

    if verifier := verifier:
        return _verified(validator, verifier=verifier)

    else:
        return validator



from asyncio import CancelledError

import pytest
from draive import AsyncQueue, ctx
from pytest import raises


class FakeException(Exception):
    pass


@pytest.mark.asyncio
async def test_fails_when_stream_fails():
    stream: AsyncQueue[int] = AsyncQueue()
    stream.enqueue(0)
    stream.finish(exception=FakeException())
    elements: int = 0
    with raises(FakeException):
        async for _ in stream:
            elements += 1

    assert elements == 1


@pytest.mark.asyncio
async def test_cancels_when_iteration_cancels():
    stream: AsyncQueue[int] = AsyncQueue()
    elements: int = 0
    with raises(CancelledError):
        ctx.cancel()
        async for _ in stream:
            elements += 1

    assert elements == 0


@pytest.mark.asyncio
async def test_ends_when_stream_ends():
    stream: AsyncQueue[int] = AsyncQueue()
    stream.finish()
    elements: int = 0
    async for _ in stream:
        elements += 1

    assert elements == 0


@pytest.mark.asyncio
async def test_buffers_values_when_not_reading():
    stream: AsyncQueue[int] = AsyncQueue()
    stream.enqueue(0)
    stream.enqueue(1)
    stream.enqueue(2)
    stream.enqueue(3)
    stream.finish()
    elements: int = 0

    async for _ in stream:
        elements += 1

    assert elements == 4


@pytest.mark.asyncio
async def test_delivers_buffer_when_streaming_fails():
    stream: AsyncQueue[int] = AsyncQueue()
    stream.enqueue(0)
    stream.enqueue(1)
    stream.enqueue(2)
    stream.enqueue(3)
    stream.finish(exception=FakeException())
    elements: int = 0

    with raises(FakeException):
        async for _ in stream:
            elements += 1

    assert elements == 4


@pytest.mark.asyncio
async def test_delivers_updates_when_sending():
    stream: AsyncQueue[int] = AsyncQueue()
    stream.enqueue(0)

    elements: list[int] = []

    async for element in stream:
        elements.append(element)
        if len(elements) < 10:
            stream.enqueue(element + 1)
        else:
            stream.finish()

    assert elements == list(range(0, 10))


@pytest.mark.asyncio
async def test_fails_when_sending_to_finished():
    stream: AsyncQueue[int] = AsyncQueue()
    stream.finish()

    with raises(RuntimeError):
        stream.enqueue(42)


@pytest.mark.asyncio
async def test_ignores_when_finishing_when_finished():
    stream: AsyncQueue[int] = AsyncQueue()
    stream.finish()
    stream.finish()  # should not raise



from collections.abc import Callable, Collection, Iterable
from typing import Any, Literal, Self, cast, final

from draive.parameters.errors import ParameterValidationError
from draive.parameters.path import ParameterPath
from draive.parameters.validation import ParameterValidationContext
from draive.utils import freeze

__all__ = [
    "ParameterRequirement",
]


@final
class ParameterRequirement[Root]:
    @classmethod
    def equal[Parameter](
        cls,
        value: Parameter,
        /,
        path: ParameterPath[Root, Parameter] | Parameter,
    ) -> Self:
        assert isinstance(  # nosec: B101
            path, ParameterPath
        ), "Prepare parameter path by using Self._.path.to.property"

        def check_equal(root: Root) -> None:
            checked: Any = cast(ParameterPath[Root, Parameter], path)(root)
            if checked != value:
                raise ParameterValidationError.invalid(
                    context=ParameterValidationContext(
                        path=cast(ParameterPath[Root, Parameter], path).components(),
                    ),
                    exception=ValueError(f"{checked} is not equal {value}"),
                )

        return cls(
            path,
            "equal",
            value,
            check=check_equal,
        )

    @classmethod
    def not_equal[Parameter](
        cls,
        value: Parameter,
        /,
        path: ParameterPath[Root, Parameter] | Parameter,
    ) -> Self:
        assert isinstance(  # nosec: B101
            path, ParameterPath
        ), "Prepare parameter path by using Self._.path.to.property"

        def check_not_equal(root: Root) -> None:
            checked: Any = cast(ParameterPath[Root, Parameter], path)(root)
            if checked == value:
                raise ParameterValidationError.invalid(
                    context=ParameterValidationContext(
                        path=cast(ParameterPath[Root, Parameter], path).components(),
                    ),
                    exception=ValueError(f"{checked} is equal {value}"),
                )

        return cls(
            path,
            "not_equal",
            value,
            check=check_not_equal,
        )

    @classmethod
    def contains[Parameter](
        cls,
        value: Parameter,
        /,
        path: ParameterPath[
            Root,
            Collection[Parameter] | tuple[Parameter, ...] | list[Parameter] | set[Parameter],
        ]
        | Collection[Parameter]
        | tuple[Parameter, ...]
        | list[Parameter]
        | set[Parameter],
    ) -> Self:
        assert isinstance(  # nosec: B101
            path, ParameterPath
        ), "Prepare parameter path by using Self._.path.to.property"

        def check_contains(root: Root) -> None:
            checked: Any = cast(ParameterPath[Root, Parameter], path)(root)
            if value not in checked:
                raise ParameterValidationError.invalid(
                    context=ParameterValidationContext(
                        path=cast(ParameterPath[Root, Parameter], path).components(),
                    ),
                    exception=ValueError(f"{checked} does not contain {value}"),
                )

        return cls(
            path,
            "contains",
            value,
            check=check_contains,
        )

    @classmethod
    def contains_any[Parameter](
        cls,
        value: Collection[Parameter],
        /,
        path: ParameterPath[
            Root,
            Collection[Parameter] | tuple[Parameter, ...] | list[Parameter] | set[Parameter],
        ]
        | Collection[Parameter]
        | tuple[Parameter, ...]
        | list[Parameter]
        | set[Parameter],
    ) -> Self:
        assert isinstance(  # nosec: B101
            path, ParameterPath
        ), "Prepare parameter path by using Self._.path.to.property"

        def check_contains_any(root: Root) -> None:
            checked: Any = cast(ParameterPath[Root, Parameter], path)(root)
            if any(element in checked for element in value):
                raise ParameterValidationError.invalid(
                    context=ParameterValidationContext(
                        path=cast(ParameterPath[Root, Parameter], path).components(),
                    ),
                    exception=ValueError(f"{checked} does not contain any of {value}"),
                )

        return cls(
            path,
            "contains_any",
            value,
            check=check_contains_any,
        )

    @classmethod
    def contained_in[Parameter](
        cls,
        value: Collection[Parameter],
        /,
        path: ParameterPath[Root, Parameter] | Parameter,
    ) -> Self:
        assert isinstance(  # nosec: B101
            path, ParameterPath
        ), "Prepare parameter path by using Self._.path.to.property"

        def check_contained_in(root: Root) -> None:
            checked: Any = cast(ParameterPath[Root, Parameter], path)(root)
            if checked not in value:
                raise ParameterValidationError.invalid(
                    context=ParameterValidationContext(
                        path=cast(ParameterPath[Root, Parameter], path).components(),
                    ),
                    exception=ValueError(f"{value} does not contain {checked}"),
                )

        return cls(
            value,
            "contained_in",
            path,
            check=check_contained_in,
        )

    def __init__(
        self,
        lhs: Any,
        operator: Literal[
            "equal",
            "not_equal",
            "contains",
            "contains_any",
            "contained_in",
            "and",
            "or",
        ],
        rhs: Any,
        check: Callable[[Root], None],
    ) -> None:
        self.lhs: Any = lhs
        self.operator: Literal[
            "equal",
            "not_equal",
            "contains",
            "contains_any",
            "contained_in",
            "and",
            "or",
        ] = operator
        self.rhs: Any = rhs
        self._check: Callable[[Root], None] = check

        freeze(self)

    def __and__(
        self,
        other: Self,
    ) -> Self:
        def check_and(root: Root) -> None:
            self.check(root)
            other.check(root)

        return self.__class__(
            self,
            "and",
            other,
            check=check_and,
        )

    def __or__(
        self,
        other: Self,
    ) -> Self:
        def check_or(root: Root) -> None:
            try:
                self.check(root)
            except ValueError:
                other.check(root)

        return self.__class__(
            self,
            "or",
            other,
            check=check_or,
        )

    def check(
        self,
        root: Root,
        /,
        *,
        raise_exception: bool = True,
    ) -> bool:
        try:
            self._check(root)
            return True

        except Exception as exc:
            if raise_exception:
                raise exc

            else:
                return False

    def filter(
        self,
        values: Iterable[Root],
    ) -> list[Root]:
        return [value for value in values if self.check(value, raise_exception=False)]



from collections.abc import Sequence
from typing import Any, Self

from draive.parameters import State
from draive.utils import MISSING, Missing

__all__ = [
    "ArgumentsTrace",
    "ExceptionTrace",
    "ResultTrace",
]


class ArgumentsTrace(State):
    if __debug__:

        @classmethod
        def of(cls, *args: Any, **kwargs: Any) -> Self:
            return cls(
                args=args if args else MISSING,
                kwargs=kwargs if kwargs else MISSING,
            )
    else:

        @classmethod
        def of(cls, *args: Any, **kwargs: Any) -> Self:
            return cls(
                args=MISSING,
                kwargs=MISSING,
            )

    args: tuple[Any, ...] | Missing
    kwargs: dict[str, Any] | Missing


class ResultTrace(State):
    @classmethod
    def of(
        cls,
        value: Any,
        /,
    ) -> Self:
        return cls(result=value)

    result: Any


class ExceptionTrace(State):
    @classmethod
    def of(
        cls,
        exception: BaseException,
        /,
    ) -> Self:
        return cls(
            name=type(exception).__qualname__,
            exception=exception,
        )

    name: str
    exception: BaseException

    def __add__(self, other: Self) -> Self:
        # the code below does not keep proper exception semantics of BaseException/Exception
        # however we are using it only for logging purposes at the moment
        # because of that merging exceptions in groups is simplified under BaseExceptionGroup
        exceptions: Sequence[BaseException]
        exception_messages: Sequence[str]
        if isinstance(self.exception, BaseExceptionGroup):
            exceptions = []
            exception_messages = []
            for exception in (*self.exception.exceptions, other.exception):  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
                exceptions.append(exception)  # pyright: ignore[reportUnknownArgumentType]
                exception_messages.append(f"{type(exception).__qualname__}:{exception}")  # pyright: ignore[reportUnknownArgumentType]

        else:
            exceptions = (self.exception, other.exception)
            exception_messages = (
                f"{type(self.exception).__qualname__}:{self.exception}",
                f"{type(other.exception).__qualname__}:{other.exception}",
            )

        return self.__class__(
            name="ExceptionGroup",
            exception=BaseExceptionGroup(
                f"Multiple errors: [{','.join(exception_messages)}]",
                exceptions,
            ),
        )



from asyncio import Task, sleep

from draive.agents.idle import IdleMonitor
from pytest import mark


@mark.asyncio
async def test_not_idle_initially():
    monitor: IdleMonitor = IdleMonitor()
    assert not monitor.idle


@mark.asyncio
async def test_becomes_idle_initially():
    monitor: IdleMonitor = IdleMonitor()

    await sleep(0)  # allow loop to handle other tasks
    assert monitor.idle


@mark.asyncio
async def test_not_idle_when_entered_task():
    monitor: IdleMonitor = IdleMonitor()

    monitor.enter_task()

    await sleep(0)  # allow loop to handle other tasks
    assert not monitor.idle


@mark.asyncio
async def test_becomes_idle_after_exiting_task():
    monitor: IdleMonitor = IdleMonitor()

    monitor.enter_task()
    await sleep(0)  # allow loop to handle other tasks

    monitor.exit_task()
    assert not monitor.idle

    await sleep(0)  # allow loop to handle other tasks

    assert monitor.idle


@mark.asyncio
async def test_not_idle_when_nested_not_idle():
    monitor: IdleMonitor = IdleMonitor()
    nested: IdleMonitor = monitor.nested()

    nested.enter_task()

    await sleep(0)  # allow loop to handle other tasks
    assert not monitor.idle


@mark.asyncio
async def test_becomes_idle_after_nested_becomes_idle():
    monitor: IdleMonitor = IdleMonitor()
    nested: IdleMonitor = monitor.nested()

    nested.enter_task()
    await sleep(0)  # allow loop to handle other tasks

    nested.exit_task()
    assert not monitor.idle

    await sleep(0)  # allow loop to handle other tasks

    assert monitor.idle


@mark.asyncio
async def test_waits_for_being_idle():
    monitor: IdleMonitor = IdleMonitor()
    monitor.enter_task()

    async def exit_task() -> None:
        await sleep(0)  # allow loop to handle other tasks
        monitor.exit_task()

    Task(exit_task())

    assert not monitor.idle
    await monitor.wait_idle()
    assert monitor.idle


@mark.asyncio
async def test_waits_for_being_idle_with_nested():
    monitor: IdleMonitor = IdleMonitor()
    nested: IdleMonitor = monitor.nested()
    nested.enter_task()

    async def exit_task() -> None:
        await sleep(0)  # allow loop to handle other tasks
        nested.exit_task()

    Task(exit_task())

    assert not monitor.idle
    await monitor.wait_idle()
    assert monitor.idle


@mark.asyncio
async def test_resets_when_entering_task():
    monitor: IdleMonitor = IdleMonitor()

    await sleep(0)  # allow loop to handle other tasks
    assert monitor.idle

    monitor.enter_task()
    await sleep(0)  # allow loop to handle other tasks
    assert not monitor.idle


@mark.asyncio
async def test_resets_when_nested_entering_task():
    monitor: IdleMonitor = IdleMonitor()
    nested: IdleMonitor = monitor.nested()

    await sleep(0)  # allow loop to handle other tasks
    assert monitor.idle

    nested.enter_task()
    await sleep(0)  # allow loop to handle other tasks
    assert not monitor.idle



from collections.abc import Iterable, Sequence
from typing import Any, Literal

from draive.generation.model.generator import ModelGeneratorDecoder
from draive.generation.model.state import ModelGeneration
from draive.instructions import Instruction
from draive.lmm import AnyTool, Toolbox
from draive.parameters import DataModel
from draive.scope import ctx
from draive.types import MultimodalContent, MultimodalContentConvertible

__all__ = [
    "generate_model",
]


async def generate_model[Generated: DataModel](  # noqa: PLR0913
    generated: type[Generated],
    /,
    *,
    instruction: Instruction | str,
    input: MultimodalContent | MultimodalContentConvertible,  # noqa: A002
    schema_injection: Literal["auto", "full", "simplified", "skip"] = "auto",
    tools: Toolbox | Sequence[AnyTool] | None = None,
    examples: Iterable[tuple[MultimodalContent | MultimodalContentConvertible, Generated]]
    | None = None,
    decoder: ModelGeneratorDecoder | None = None,
    **extra: Any,
) -> Generated:
    return await ctx.state(ModelGeneration).generate(
        generated,
        instruction=instruction,
        input=input,
        schema_injection=schema_injection,
        tools=tools,
        examples=examples,
        decoder=decoder,
        **extra,
    )



from collections.abc import AsyncIterable
from datetime import datetime
from typing import Literal, Self
from uuid import uuid4

from draive.lmm import ToolStatus
from draive.parameters import DataModel, Field
from draive.types import (
    LMMCompletion,
    LMMContextElement,
    LMMInput,
    MultimodalContent,
    MultimodalContentConvertible,
)

__all__ = [
    "ConversationMessage",
    "ConversationMessageChunk",
    "ConversationResponseStream",
]


class ConversationMessage(DataModel):
    @classmethod
    def user(
        cls,
        content: MultimodalContent | MultimodalContentConvertible,
        identifier: str | None = None,
        author: str | None = None,
        created: datetime | None = None,
    ) -> Self:
        return cls(
            identifier=identifier or uuid4().hex,
            role="user",
            author=author,
            created=created,
            content=MultimodalContent.of(content),
        )

    @classmethod
    def model(
        cls,
        content: MultimodalContent | MultimodalContentConvertible,
        identifier: str | None = None,
        author: str | None = None,
        created: datetime | None = None,
    ) -> Self:
        return cls(
            identifier=identifier or uuid4().hex,
            role="model",
            author=author,
            created=created,
            content=MultimodalContent.of(content),
        )

    identifier: str = Field(default_factory=lambda: uuid4().hex)
    role: Literal["user", "model"]
    author: str | None = None
    created: datetime | None = None
    content: MultimodalContent

    def as_lmm_context_element(self) -> LMMContextElement:
        match self.role:
            case "user":
                return LMMInput.of(self.content)

            case "model":
                return LMMCompletion.of(self.content)

    def __bool__(self) -> bool:
        return bool(self.content)


class ConversationMessageChunk(DataModel):
    identifier: str
    content: MultimodalContent

    def __bool__(self) -> bool:
        return bool(self.content)


ConversationResponseStream = AsyncIterable[ConversationMessageChunk | ToolStatus]



from collections.abc import Sequence
from typing import Any

from draive.embedding.embedded import Embedded
from draive.embedding.state import ImageEmbedding, TextEmbedding
from draive.scope import ctx

__all__ = [
    "embed_text",
    "embed_texts",
    "embed_image",
    "embed_images",
]


async def embed_text(
    text: str,
    /,
    **extra: Any,
) -> Embedded[str]:
    return (
        await ctx.state(TextEmbedding).embed(
            [text],
            **extra,
        )
    )[0]


async def embed_texts(
    texts: Sequence[str],
    /,
    **extra: Any,
) -> list[Embedded[str]]:
    return await ctx.state(TextEmbedding).embed(
        texts,
        **extra,
    )


async def embed_image(
    image: bytes,
    /,
    **extra: Any,
) -> Embedded[bytes]:
    return (
        await ctx.state(ImageEmbedding).embed(
            [image],
            **extra,
        )
    )[0]


async def embed_images(
    images: Sequence[bytes],
    /,
    **extra: Any,
) -> list[Embedded[bytes]]:
    return await ctx.state(ImageEmbedding).embed(
        images,
        **extra,
    )



from draive.types.audio import AudioBase64Content, AudioContent, AudioURLContent
from draive.types.errors import RateLimitError
from draive.types.frozenlist import frozenlist
from draive.types.image import ImageBase64Content, ImageContent, ImageURLContent
from draive.types.json import JSON
from draive.types.lmm import (
    LMMCompletion,
    LMMCompletionChunk,
    LMMContextElement,
    LMMInput,
    LMMOutput,
    LMMOutputStream,
    LMMOutputStreamChunk,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponse,
)
from draive.types.memory import BasicMemory, Memory
from draive.types.multimodal import (
    MultimodalContent,
    MultimodalContentConvertible,
    MultimodalContentElement,
)
from draive.types.text import TextContent
from draive.types.video import VideoBase64Content, VideoContent, VideoURLContent

__all__ = [
    "AudioBase64Content",
    "AudioContent",
    "AudioURLContent",
    "BasicMemory",
    "frozenlist",
    "ImageBase64Content",
    "ImageContent",
    "ImageURLContent",
    "JSON",
    "LMMCompletion",
    "LMMCompletionChunk",
    "LMMContextElement",
    "LMMInput",
    "LMMOutput",
    "LMMOutputStream",
    "LMMOutputStreamChunk",
    "LMMToolRequest",
    "LMMToolRequests",
    "LMMToolResponse",
    "Memory",
    "MultimodalContent",
    "MultimodalContentConvertible",
    "MultimodalContentElement",
    "RateLimitError",
    "TextContent",
    "VideoBase64Content",
    "VideoContent",
    "VideoURLContent",
]
